# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ctypes
from dataclasses import dataclass
from typing import Any

import torch
from torch.distributed import ReduceOp

from cosmos_rl.utils.logging import logger


# === export types and functions from nccl to Python ===
# for the original nccl definition, please check
# https://github.com/NVIDIA/nccl/blob/master/src/nccl.h.in

ncclResult_t = ctypes.c_int
ncclComm_t = ctypes.c_void_p


class ncclUniqueId(ctypes.Structure):
    _fields_ = [("internal", ctypes.c_byte * 128)]


cudaStream_t = ctypes.c_void_p
buffer_type = ctypes.c_void_p

ncclDataType_t = ctypes.c_int


# --- NCCL config struct (complete v22700) ---
class ncclConfig_t(ctypes.Structure):
    _fields_ = [
        ("size", ctypes.c_size_t),  # sizeof(ncclConfig_t)
        ("magic", ctypes.c_uint),  # constant magic
        ("version", ctypes.c_uint),  # NCCL version code, e.g. 22703
        ("blocking", ctypes.c_int),  # whether operations are blocking (0 / 1)
        ("cgaClusterSize", ctypes.c_int),
        ("minCTAs", ctypes.c_int),
        ("maxCTAs", ctypes.c_int),
        ("netName", ctypes.c_char_p),
        ("splitShare", ctypes.c_int),
        ("trafficClass", ctypes.c_int),
        ("commName", ctypes.c_char_p),
        ("collnetEnable", ctypes.c_int),
        ("CTAPolicy", ctypes.c_int),
        ("shrinkShare", ctypes.c_int),
        ("nvlsCTAs", ctypes.c_int),
    ]


class ncclDataTypeEnum:
    ncclInt8 = 0
    ncclChar = 0
    ncclUint8 = 1
    ncclInt32 = 2
    ncclInt = 2
    ncclUint32 = 3
    ncclInt64 = 4
    ncclUint64 = 5
    ncclFloat16 = 6
    ncclHalf = 6
    ncclFloat32 = 7
    ncclFloat = 7
    ncclFloat64 = 8
    ncclDouble = 8
    ncclBfloat16 = 9
    ncclFloat8e4m3 = 10
    ncclFloat8e5m2 = 11

    @classmethod
    def from_torch(cls, dtype: torch.dtype) -> int:
        if dtype == torch.int8:
            return cls.ncclInt8
        if dtype == torch.uint8:
            return cls.ncclUint8
        if dtype == torch.int32:
            return cls.ncclInt32
        if dtype == torch.int64:
            return cls.ncclInt64
        if dtype == torch.float16:
            return cls.ncclFloat16
        if dtype == torch.float32:
            return cls.ncclFloat32
        if dtype == torch.float64:
            return cls.ncclFloat64
        if dtype == torch.bfloat16:
            return cls.ncclBfloat16
        if dtype == torch.float8_e4m3fn:
            return cls.ncclFloat8e4m3
        if dtype == torch.float8_e5m2:
            return cls.ncclFloat8e5m2
        raise ValueError(f"Unsupported dtype: {dtype}")


ncclRedOp_t = ctypes.c_int


class ncclRedOpTypeEnum:
    ncclSum = 0
    ncclProd = 1
    ncclMax = 2
    ncclMin = 3
    ncclAvg = 4
    ncclNumOps = 5

    @classmethod
    def from_torch(cls, op: ReduceOp) -> int:
        if op == ReduceOp.SUM:
            return cls.ncclSum
        if op == ReduceOp.PRODUCT:
            return cls.ncclProd
        if op == ReduceOp.MAX:
            return cls.ncclMax
        if op == ReduceOp.MIN:
            return cls.ncclMin
        if op == ReduceOp.AVG:
            return cls.ncclAvg
        raise ValueError(f"Unsupported op: {op}")


@dataclass
class Function:
    name: str
    restype: Any
    argtypes: list[Any]


class NCCLLibrary:
    # names of optional functions (absence tolerated)
    optional_functions = {"ncclCommInitRankConfig"}
    exported_functions = [
        # const char* ncclGetErrorString(ncclResult_t result)
        Function("ncclGetErrorString", ctypes.c_char_p, [ncclResult_t]),
        # ncclResult_t  ncclGetVersion(int *version);
        Function("ncclGetVersion", ncclResult_t, [ctypes.POINTER(ctypes.c_int)]),
        # ncclResult_t ncclGetUniqueId(ncclUniqueId* uniqueId);
        Function("ncclGetUniqueId", ncclResult_t, [ctypes.POINTER(ncclUniqueId)]),
        # ncclResult_t  ncclCommInitRank(
        #   ncclComm_t* comm, int nranks, ncclUniqueId commId, int rank);
        # note that ncclComm_t is a pointer type, so the first argument
        # is a pointer to a pointer
        Function(
            "ncclCommInitRank",
            ncclResult_t,
            [ctypes.POINTER(ncclComm_t), ctypes.c_int, ncclUniqueId, ctypes.c_int],
        ),
        # ncclResult_t  ncclAllReduce(
        #   const void* sendbuff, void* recvbuff, size_t count,
        #   ncclDataType_t datatype, ncclRedOp_t op, ncclComm_t comm,
        #   cudaStream_t stream);
        # note that cudaStream_t is a pointer type, so the last argument
        # is a pointer
        Function(
            "ncclAllReduce",
            ncclResult_t,
            [
                buffer_type,
                buffer_type,
                ctypes.c_size_t,
                ncclDataType_t,
                ncclRedOp_t,
                ncclComm_t,
                cudaStream_t,
            ],
        ),
        # ncclResult_t  ncclAllGather(
        #   const void* sendbuff, void* recvbuff, size_t count,
        #   ncclDataType_t datatype, ncclComm_t comm,
        #   cudaStream_t stream);
        # note that cudaStream_t is a pointer type, so the last argument
        # is a pointer
        Function(
            "ncclAllGather",
            ncclResult_t,
            [
                buffer_type,
                buffer_type,
                ctypes.c_size_t,
                ncclDataType_t,
                ncclComm_t,
                cudaStream_t,
            ],
        ),
        # ncclResult_t  ncclReduceScatter(
        #   const void* sendbuff, void* recvbuff, size_t count,
        #   ncclDataType_t datatype, ncclRedOp_t op, ncclComm_t comm,
        #   cudaStream_t stream);
        # note that cudaStream_t is a pointer type, so the last argument
        # is a pointer
        Function(
            "ncclReduceScatter",
            ncclResult_t,
            [
                buffer_type,
                buffer_type,
                ctypes.c_size_t,
                ncclDataType_t,
                ncclRedOp_t,
                ncclComm_t,
                cudaStream_t,
            ],
        ),
        # ncclResult_t  ncclSend(
        #   const void* sendbuff, size_t count, ncclDataType_t datatype,
        #   int dest, ncclComm_t comm, cudaStream_t stream);
        Function(
            "ncclSend",
            ncclResult_t,
            [
                buffer_type,
                ctypes.c_size_t,
                ncclDataType_t,
                ctypes.c_int,
                ncclComm_t,
                cudaStream_t,
            ],
        ),
        # ncclResult_t  ncclRecv(
        #   void* recvbuff, size_t count, ncclDataType_t datatype,
        #   int src, ncclComm_t comm, cudaStream_t stream);
        Function(
            "ncclRecv",
            ncclResult_t,
            [
                buffer_type,
                ctypes.c_size_t,
                ncclDataType_t,
                ctypes.c_int,
                ncclComm_t,
                cudaStream_t,
            ],
        ),
        # ncclResult_t ncclBroadcast(
        #   const void* sendbuff, void* recvbuff, size_t count,
        #   ncclDataType_t datatype, int root, ncclComm_t comm,
        #   cudaStream_t stream);
        Function(
            "ncclBroadcast",
            ncclResult_t,
            [
                buffer_type,
                buffer_type,
                ctypes.c_size_t,
                ncclDataType_t,
                ctypes.c_int,
                ncclComm_t,
                cudaStream_t,
            ],
        ),
        # be cautious! this is a collective call, it will block until all
        # processes in the communicator have called this function.
        # because Python object destruction can happen in random order,
        # it is better not to call it at all.
        # ncclResult_t  ncclCommDestroy(ncclComm_t comm);
        Function("ncclCommDestroy", ncclResult_t, [ncclComm_t]),
        # NCCL async error query & abort (for enqueue-timeout protection)
        Function(
            "ncclCommGetAsyncError",
            ncclResult_t,
            [ncclComm_t, ctypes.POINTER(ncclResult_t)],
        ),
        Function("ncclCommAbort", ncclResult_t, [ncclComm_t]),
        # ncclResult_t  ncclCommInitRankConfig(
        #   ncclComm_t* comm, int nranks, ncclUniqueId commId, int rank,
        #   ncclConfig_t* config);
        Function(
            "ncclCommInitRankConfig",
            ncclResult_t,
            [
                ctypes.POINTER(ncclComm_t),
                ctypes.c_int,
                ncclUniqueId,
                ctypes.c_int,
                ctypes.POINTER(ncclConfig_t),
            ],
        ),
        # ncclResult_t ncclGroupStart()
        Function(
            "ncclGroupStart",
            ncclResult_t,
            [],
        ),
        # ncclResult_t ncclGroupEnd()
        Function(
            "ncclGroupEnd",
            ncclResult_t,
            [],
        ),
    ]

    # class attribute to store the mapping from the path to the library
    # to avoid loading the same library multiple times
    path_to_library_cache: dict[str, Any] = {}

    # class attribute to store the mapping from library path
    #  to the corresponding functions
    path_to_funcs_mapping: dict[str, dict[str, Any]] = {}

    def __init__(self, so_file: str):
        try:
            if so_file not in NCCLLibrary.path_to_library_cache:
                lib = ctypes.CDLL(so_file)
                NCCLLibrary.path_to_library_cache[so_file] = lib
            self.lib = NCCLLibrary.path_to_library_cache[so_file]
        except Exception:
            logger.error(f"Failed to load so file: {so_file} from NCCL library. ")
            raise

        if so_file not in NCCLLibrary.path_to_funcs_mapping:
            _funcs: dict[str, Any] = {}
            for func in NCCLLibrary.exported_functions:
                try:
                    f = getattr(self.lib, func.name)
                except AttributeError:
                    if func.name in NCCLLibrary.optional_functions:
                        logger.warning(
                            f"Optional NCCL symbol {func.name} not found; falling back to default behavior."
                        )
                        _funcs[func.name] = None
                        continue
                    raise
                f.restype = func.restype
                f.argtypes = func.argtypes
                _funcs[func.name] = f
            NCCLLibrary.path_to_funcs_mapping[so_file] = _funcs
        self._funcs = NCCLLibrary.path_to_funcs_mapping[so_file]

    def ncclGetErrorString(self, result: ncclResult_t) -> str:
        return self._funcs["ncclGetErrorString"](result).decode("utf-8")

    def NCCL_CHECK(self, result: ncclResult_t) -> None:
        """Raise RuntimeError if *result* is an error code."""
        if not ncclResultEnum.is_ok(int(result)):
            error_str = self.ncclGetErrorString(result)
            raise RuntimeError(f"NCCL error: {error_str}")

    def ncclGetVersion(self) -> str:
        version = ctypes.c_int()
        self.NCCL_CHECK(self._funcs["ncclGetVersion"](ctypes.byref(version)))
        version_str = str(version.value)
        # something like 21903 --> "2.19.3"
        major = version_str[0].lstrip("0")
        minor = version_str[1:3].lstrip("0")
        patch = version_str[3:].lstrip("0")
        return f"{major}.{minor}.{patch}"

    def ncclGetUniqueId(self) -> ncclUniqueId:
        unique_id = ncclUniqueId()
        self.NCCL_CHECK(self._funcs["ncclGetUniqueId"](ctypes.byref(unique_id)))
        return unique_id

    def ncclCommInitRank(
        self, world_size: int, unique_id: ncclUniqueId, rank: int
    ) -> ncclComm_t:
        comm = ncclComm_t()
        self.NCCL_CHECK(
            self._funcs["ncclCommInitRank"](
                ctypes.byref(comm), world_size, unique_id, rank
            )
        )
        return comm

    def ncclCommInitRankConfig(
        self,
        world_size: int,
        unique_id: ncclUniqueId,
        rank: int,
        blocking: int = 0,
    ) -> ncclComm_t:
        """Python wrapper for ncclCommInitRankConfig with *blocking* flag preset.

        Currently only exposes the *blocking* field (0 = non-blocking NCCL launch).
        Additional ncclConfig_t fields are kept at their default zeros for
        simplicity.
        """
        comm = ncclComm_t()
        config = ncclConfig_t()
        # Fill mandatory header fields according to NCCL_CONFIG_INITIALIZER
        NCCL_CONFIG_MAGIC = 0xCAFEBEEF
        # Try to retrieve NCCL version code (e.g., 22703) from library
        try:
            ver_str = self.ncclGetVersion()  # e.g. "2.27.3"
            major, minor, patch = (int(v) for v in ver_str.split("."))
            version_code = (major * 10000) + (minor * 100) + patch
        except Exception:
            # Fallback to hard-coded current header version
            version_code = 22703

        INT_MIN = -2147483648  # NCCL_CONFIG_UNDEF_INT

        config.size = ctypes.sizeof(ncclConfig_t)
        config.magic = NCCL_CONFIG_MAGIC
        config.version = version_code

        # Initialize all integer fields to UNDEF, pointer fields to NULL
        config.blocking = blocking if blocking in (0, 1) else INT_MIN
        config.cgaClusterSize = INT_MIN
        config.minCTAs = INT_MIN
        config.maxCTAs = INT_MIN
        config.netName = None
        config.splitShare = INT_MIN
        config.trafficClass = INT_MIN
        config.commName = None
        config.collnetEnable = INT_MIN
        config.CTAPolicy = INT_MIN
        config.shrinkShare = INT_MIN
        config.nvlsCTAs = INT_MIN

        fn = self._funcs.get("ncclCommInitRankConfig")
        if fn is None:
            logger.debug(
                "ncclCommInitRankConfig symbol missing – falling back to ncclCommInitRank"
            )
            return self.ncclCommInitRank(world_size, unique_id, rank)

        ret = fn(
            ctypes.byref(comm),
            world_size,
            unique_id,
            rank,
            ctypes.byref(config),
        )
        if not ncclResultEnum.is_ok(int(ret)):
            self.NCCL_CHECK(ret)
        return comm

    def ncclAllReduce(
        self,
        sendbuff: buffer_type,
        recvbuff: buffer_type,
        count: int,
        datatype: int,
        op: int,
        comm: ncclComm_t,
        stream: cudaStream_t,
    ) -> None:
        # `datatype` actually should be `ncclDataType_t`
        # and `op` should be `ncclRedOp_t`
        # both are aliases of `ctypes.c_int`
        # when we pass int to a function, it will be converted to `ctypes.c_int`
        # by ctypes automatically
        self.NCCL_CHECK(
            self._funcs["ncclAllReduce"](
                sendbuff, recvbuff, count, datatype, op, comm, stream
            )
        )

    def ncclReduceScatter(
        self,
        sendbuff: buffer_type,
        recvbuff: buffer_type,
        count: int,
        datatype: int,
        op: int,
        comm: ncclComm_t,
        stream: cudaStream_t,
    ) -> None:
        # `datatype` actually should be `ncclDataType_t`
        # and `op` should be `ncclRedOp_t`
        # both are aliases of `ctypes.c_int`
        # when we pass int to a function, it will be converted to `ctypes.c_int`
        # by ctypes automatically
        self.NCCL_CHECK(
            self._funcs["ncclReduceScatter"](
                sendbuff, recvbuff, count, datatype, op, comm, stream
            )
        )

    def ncclAllGather(
        self,
        sendbuff: buffer_type,
        recvbuff: buffer_type,
        count: int,
        datatype: int,
        comm: ncclComm_t,
        stream: cudaStream_t,
    ) -> None:
        # `datatype` actually should be `ncclDataType_t`
        # which is an aliases of `ctypes.c_int`
        # when we pass int to a function, it will be converted to `ctypes.c_int`
        # by ctypes automatically
        self.NCCL_CHECK(
            self._funcs["ncclAllGather"](
                sendbuff, recvbuff, count, datatype, comm, stream
            )
        )

    def ncclSend(
        self,
        sendbuff: buffer_type,
        count: int,
        datatype: int,
        dest: int,
        comm: ncclComm_t,
        stream: cudaStream_t,
    ) -> None:
        self.NCCL_CHECK(
            self._funcs["ncclSend"](sendbuff, count, datatype, dest, comm, stream)
        )

    def ncclRecv(
        self,
        recvbuff: buffer_type,
        count: int,
        datatype: int,
        src: int,
        comm: ncclComm_t,
        stream: cudaStream_t,
    ) -> None:
        self.NCCL_CHECK(
            self._funcs["ncclRecv"](recvbuff, count, datatype, src, comm, stream)
        )

    def ncclGroupStart(self) -> None:
        self.NCCL_CHECK(self._funcs["ncclGroupStart"]())

    def ncclGroupEnd(self) -> None:
        self.NCCL_CHECK(self._funcs["ncclGroupEnd"]())

    def ncclBroadcast(
        self,
        sendbuff: buffer_type,
        recvbuff: buffer_type,
        count: int,
        datatype: int,
        root: int,
        comm: ncclComm_t,
        stream: cudaStream_t,
    ) -> None:
        self.NCCL_CHECK(
            self._funcs["ncclBroadcast"](
                sendbuff, recvbuff, count, datatype, root, comm, stream
            )
        )

    def ncclCommDestroy(self, comm: ncclComm_t) -> None:
        self.NCCL_CHECK(self._funcs["ncclCommDestroy"](comm))

    # ------------------ new helpers for HA ------------------

    def ncclCommGetAsyncError(self, comm: ncclComm_t) -> int:
        err = ncclResult_t()
        self._funcs["ncclCommGetAsyncError"](comm, ctypes.byref(err))
        return int(err.value)

    def ncclCommAbort(self, comm: ncclComm_t) -> None:
        # abort can return error itself; ignore it to avoid masking original issue
        self._funcs["ncclCommAbort"](comm)


class ncclResultEnum:
    """Enumeration of NCCL result codes copied from nccl.h."""

    ncclSuccess = 0
    ncclUnhandledCudaError = 1
    ncclSystemError = 2
    ncclInternalError = 3
    ncclInvalidArgument = 4
    ncclInvalidUsage = 5
    ncclRemoteError = 6
    ncclInProgress = 7

    @staticmethod
    def is_ok(result: int) -> bool:
        """Return True if *result* represents a non-error condition."""
        return result in (
            ncclResultEnum.ncclSuccess,
            ncclResultEnum.ncclInProgress,
        )


__all__ = [
    "NCCLLibrary",
    "ncclDataTypeEnum",
    "ncclRedOpTypeEnum",
    "ncclResultEnum",
    "ncclUniqueId",
    "ncclComm_t",
    "cudaStream_t",
    "buffer_type",
    "ncclConfig_t",
]



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from torch.distributed.device_mesh import DeviceMesh, init_device_mesh
from cosmos_rl.utils.logging import logger
from cosmos_rl.policy.config import ParallelismConfig
import contextlib
from typing import Generator, Optional, List
import torch
import numpy
import os


def train_context(enable_compiled_autograd: bool):
    @contextlib.contextmanager
    def context(cp_context: Optional[Generator[None, None, None]] = None):
        with contextlib.ExitStack() as stack:
            if enable_compiled_autograd:
                stack.enter_context(
                    torch._dynamo.utils.maybe_enable_compiled_autograd(True)
                )

            if cp_context is not None:
                from torch.nn.attention import sdpa_kernel, SDPBackend

                stack.enter_context(
                    sdpa_kernel(
                        [
                            SDPBackend.FLASH_ATTENTION,
                            SDPBackend.EFFICIENT_ATTENTION,
                            SDPBackend.CUDNN_ATTENTION,
                        ]
                    )
                )
                stack.enter_context(cp_context)

            yield

    return context


def unshard_context_parallel_output(
    cp_mesh: DeviceMesh, buffers: List[torch.Tensor], cp_seq_dims: List[int]
) -> torch.Tensor:
    try:
        from torch.distributed.tensor.experimental._attention import (
            context_parallel_unshard,
        )
    except ImportError:
        print(
            f"PyTorch version {torch.__version__} does not include the experimental "
            "Context Parallel API. Please update to a newer version."
        )

    return context_parallel_unshard(cp_mesh, buffers, cp_seq_dims)


@dataclass
class ParallelDims:
    dp_replicate: int
    dp_shard: int
    cp: int
    tp: int
    pp: int
    world_size: int
    pp_dynamic_shape: bool

    @staticmethod
    def from_config(parallesim_config: ParallelismConfig):
        return ParallelDims(
            dp_replicate=parallesim_config.dp_replicate_size,
            dp_shard=parallesim_config.dp_shard_size,
            cp=parallesim_config.cp_size,
            tp=parallesim_config.tp_size,
            pp=parallesim_config.pp_size,
            world_size=parallesim_config.world_size,
            pp_dynamic_shape=parallesim_config.pp_dynamic_shape,
        )

    @staticmethod
    def from_config_for_analysis(parallesim_config: ParallelismConfig, world_size: int):
        return ParallelDims(
            dp_replicate=parallesim_config.dp_replicate_size,
            dp_shard=parallesim_config.dp_shard_size,
            cp=parallesim_config.cp_size,
            tp=parallesim_config.tp_size,
            pp=parallesim_config.pp_size,
            world_size=world_size,
            pp_dynamic_shape=parallesim_config.pp_dynamic_shape,
        )

    def __post_init__(self):
        self._validate()
        self.build_mesh_info()

    def _validate(self):
        dp_replicate, dp_shard, cp, tp, pp = (
            self.dp_replicate,
            self.dp_shard,
            self.cp,
            self.tp,
            self.pp,
        )
        for d in (dp_replicate, cp, tp, pp):
            assert d >= 1, "Parallelism degree should be >= 1, except for dp_shard"
        assert dp_shard == -1 or dp_shard >= 1, " dp_shard must be -1 or >=1."
        if dp_shard < 0:
            self.dp_shard = dp_shard = self.world_size // (dp_replicate * cp * tp * pp)
        assert (
            dp_shard >= 1
        ), f"dp_shard of size {dp_shard} is not valid, should be equal or greater than 1"

        assert dp_replicate * dp_shard * cp * tp * pp == self.world_size, (
            f"Invalid parallel dims: dp_replicate({dp_replicate}) * dp_shard({dp_shard}) * "
            f"cp({cp}) * tp({tp}) * pp({pp}) != WORLD_SIZE({self.world_size})"
        )

    def build_mesh(self, device_type: str) -> DeviceMesh:
        dims = []
        names = []
        for d, name in zip(
            [self.pp, self.dp_replicate, self.dp_shard, self.cp, self.tp],
            [
                "pp",
                "dp_replicate",
                "dp_shard",
                "cp",
                "tp",
            ],  # reverse order to apply N-dim prallel.
        ):
            if d > 1:
                dims.append(d)
                names.append(name)
        return self._build_mesh(device_type, dims, names)

    def _build_mesh(
        self,
        device_type: str,
        dims: list[int],
        names: list[str],
    ) -> DeviceMesh:
        logger.info(f"Building {len(dims)}-D device mesh with {names}, {dims}")
        names = tuple(names)
        mesh = init_device_mesh(device_type, dims, mesh_dim_names=names)

        # Create all the submesh here to ensure all required process groups are
        # initialized:
        # Mesh for data loading (no communication on this mesh)
        dp_mesh_dim_names = []
        # Mesh for param sharding
        dp_shard_cp_mesh_dim_names = []
        # Mesh useful for TP-merged FSDP
        dp_cp_tp_mesh_dim_names = []
        # Mesh for loss all-reduce
        dp_cp_mesh_dim_names = []

        if self.dp_replicate_enabled:
            dp_mesh_dim_names.append("dp_replicate")
            dp_cp_mesh_dim_names.append("dp_replicate")
            dp_cp_tp_mesh_dim_names.append("dp_replicate")
        if self.dp_shard_enabled:
            dp_mesh_dim_names.append("dp_shard")
            dp_shard_cp_mesh_dim_names.append("dp_shard")
            dp_cp_tp_mesh_dim_names.append("dp_shard")
            dp_cp_mesh_dim_names.append("dp_shard")
        if self.cp_enabled:
            dp_shard_cp_mesh_dim_names.append("cp")
            dp_cp_tp_mesh_dim_names.append("cp")
            dp_cp_mesh_dim_names.append("cp")
        if self.tp_enabled:
            dp_cp_tp_mesh_dim_names.append("tp")

        if dp_mesh_dim_names != []:
            mesh[tuple(dp_mesh_dim_names)]._flatten(mesh_dim_name="dp")
        if dp_shard_cp_mesh_dim_names != []:
            mesh[tuple(dp_shard_cp_mesh_dim_names)]._flatten(
                mesh_dim_name="dp_shard_cp"
            )
        if dp_cp_tp_mesh_dim_names != []:
            mesh[tuple(dp_cp_tp_mesh_dim_names)]._flatten(mesh_dim_name="dp_cp_tp")
        if dp_cp_mesh_dim_names != []:
            mesh[tuple(dp_cp_mesh_dim_names)]._flatten(mesh_dim_name="dp_cp")

        self.mesh = mesh
        return mesh

    def get_rank_in_dim(self, mesh_dim_name: str, global_rank: int) -> int:
        if hasattr(self, "full_rank_info"):
            if mesh_dim_name in self.full_rank_info[global_rank]:
                return self.full_rank_info[global_rank][mesh_dim_name]
            else:
                raise ValueError(f"Mesh dim {mesh_dim_name} not found in rank info.")
        else:
            raise ValueError(
                "full_rank_info is not set. Please call build_mesh() first."
            )

    def get_size_in_dim(self, mesh_dim_name: str) -> int:
        if hasattr(self, "full_world_size_info"):
            if mesh_dim_name in self.full_world_size_info:
                return self.full_world_size_info[mesh_dim_name]
            else:
                raise ValueError(
                    f"Mesh dim {mesh_dim_name} not found in world size info."
                )
        else:
            try:
                return self.mesh.get_group(mesh_dim_name).size()
            except Exception:
                pass
            return 1

    @property
    def dp_enabled(self):
        return self.dp_replicate > 1 or self.dp_shard > 1

    @property
    def dp_replicate_enabled(self):
        return self.dp_replicate > 1

    @property
    def dp_shard_enabled(self):
        return self.dp_shard > 1

    @property
    def cp_enabled(self):
        return self.cp > 1

    @property
    def tp_enabled(self):
        return self.tp > 1

    @property
    def pp_enabled(self):
        return self.pp > 1

    @property
    def pp_dynamic_shape_enabled(self):
        return self.pp > 1 and self.pp_dynamic_shape

    def non_data_parallel_size(self):
        return self.cp * self.tp * self.pp

    @property
    def dp_replicate_coord(self):
        if not self.dp_replicate_enabled:
            return 0, 1
        return (self.mesh.get_local_rank(mesh_dim="dp_replicate"), self.dp_replicate)

    @property
    def tp_coord(self):
        if not self.tp_enabled:
            return 0, 1
        return (self.mesh.get_local_rank(mesh_dim="tp"), self.tp)

    @property
    def pp_coord(self):
        if not self.pp_enabled:
            return 0, 1
        return (self.mesh.get_local_rank(mesh_dim="pp"), self.pp)

    @property
    def dp_shard_coord(self):
        if not self.dp_shard_enabled:
            return 0, 1
        return (self.mesh.get_local_rank(mesh_dim="dp_shard"), self.dp_shard)

    @property
    def cp_coord(self):
        if not self.cp_enabled:
            return 0, 1
        return (self.mesh.get_local_rank(mesh_dim="cp"), self.cp)

    @property
    def dp_shard_cp_coord(self):
        if not self.dp_shard_enabled and not self.cp_enabled:
            return 0, 1
        else:
            return self.mesh[tuple(("dp_shard_cp",))].get_local_rank(), self.mesh[
                tuple(("dp_shard_cp",))
            ].size()

    def build_mesh_info(self):
        dims = ["pp", "dp_replicate", "dp_shard", "cp", "tp"]
        dim_paras = {
            "pp": self.pp,
            "dp_replicate": self.dp_replicate,
            "dp_shard": self.dp_shard,
            "cp": self.cp,
            "tp": self.tp,
        }
        info = [{} for i in range(self.world_size)]
        meshes = [range(self.world_size)]
        for dim in dims:
            new_meshes = []
            for m in meshes:
                for r, arr in enumerate(numpy.array_split(m, dim_paras[dim])):
                    for d in list(arr):
                        if d in m:
                            info[d][dim] = r
                    new_meshes.append(list(arr))
            meshes = new_meshes
        # Note: full_rank_info will record the rank in each dimension for a global rank/device.
        # e.g: [{'pp': 0, 'dp_replicate': 0, 'dp_shard': 0, 'cp': 0, 'tp': 0, 'dp_shard_cp': 0, 'dp': 0},
        # {'pp': 0, 'dp_replicate': 0, 'dp_shard': 0, 'cp': 0, 'tp': 1, 'dp_shard_cp': 0, 'dp': 0}]
        self.full_rank_info = info
        self.full_world_size_info = dim_paras
        self.full_world_size_info["dp_shard_cp"] = self.dp_shard * self.cp
        self.full_world_size_info["dp"] = self.dp_replicate * self.dp_shard
        self.full_world_size_info["dp_cp_tp"] = (
            self.dp_replicate * self.dp_shard * self.cp * self.tp
        )

        for i in range(self.world_size):
            self.full_rank_info[i]["dp_cp_tp"] = (
                self.full_rank_info[i]["dp_replicate"]
                * self.dp_shard
                * self.cp
                * self.tp
                + self.full_rank_info[i]["dp_shard"] * self.cp * self.tp
                + self.full_rank_info[i]["cp"] * self.tp
                + self.full_rank_info[i]["tp"]
            )
            self.full_rank_info[i]["dp_shard_cp"] = (
                self.full_rank_info[i]["dp_shard"] * self.cp
                + self.full_rank_info[i]["cp"]
            )
            self.full_rank_info[i]["dp"] = (
                self.full_rank_info[i]["dp_replicate"] * self.dp_shard
                + self.full_rank_info[i]["dp_shard"]
            )

        self.global_rank = int(os.environ.get("RANK", 0))
        # logger.info(f"Full rank info: {self.full_rank_info}")



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import torch

from cosmos_rl.utils.parallelism import ParallelDims
from cosmos_rl.policy.config import Config as CosmosConfig
from cosmos_rl.comm.base import CommMixin
from cosmos_rl.dispatcher.protocol import Role
import cosmos_rl.utils.util as util
from transformers import AutoTokenizer


class RolloutWorkerBase(CommMixin):
    def __init__(self, config: CosmosConfig, parallel_dims: ParallelDims) -> None:
        super().__init__()
        self.config = config
        self.role = Role.ROLLOUT
        self.parallel_dims = parallel_dims
        self.local_rank = int(os.environ.get("LOCAL_RANK", 0))  # rank in the node
        self.global_rank = int(os.environ.get("RANK", 0))  # rank in replica
        self.world_size = int(os.environ.get("WORLD_SIZE", 1))
        self.device = torch.device(f"cuda:{self.local_rank}")
        torch.cuda.set_device(self.device)
        self.tokenizer = util.retry(AutoTokenizer.from_pretrained)(
            config.policy.model_name_or_path
        )
        # Initialize the communication to controller.
        self.init_comm()
        self.init_redis()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import concurrent.futures as futures
from typing import List
import os
import torch
from functools import partial
import requests

from cosmos_rl.utils.logging import logger
from cosmos_rl.policy.config import Config as CosmosConfig
from cosmos_rl.utils.parallelism import ParallelDims
from cosmos_rl.utils.network_util import make_request_with_retry
from cosmos_rl.utils.checkpoint import upload_folder_to_s3
from cosmos_rl.utils import constant


class CosmosProfiler:
    def __init__(
        self,
        config: CosmosConfig,
        parallel_dims: ParallelDims,
        replica_name: str,
        alternative_urls: List[str],
    ):
        self.config = config
        self.replica_name = replica_name
        self.enable_profile = config.profiler.enable_profiler
        self.global_rank = parallel_dims.global_rank
        self.initialized = False
        self.is_started = False
        self.is_finished = False
        self.profiled_times = 0
        self.profiler = None
        self.thread_pool = None
        self.alternative_urls = alternative_urls
        # We do not want the timestamp part of output dir for profiling data.
        output_dir = os.path.dirname(config.train.output_dir)
        self.output_dir = os.path.join(
            output_dir, "profile_trace", f"{self.replica_name}_{self.global_rank}"
        )
        self.thread_pool = None

        # static profiler config
        self.wait_steps = config.profiler.sub_profiler_config.wait_steps
        self.warmup_steps = config.profiler.sub_profiler_config.warmup_steps
        self.active_steps = config.profiler.sub_profiler_config.active_steps
        if len(config.profiler.sub_profiler_config.rank_filter) == 0:
            self.rank_filter = [0]
        else:
            self.rank_filter = config.profiler.sub_profiler_config.rank_filter
        self.record_shape = config.profiler.sub_profiler_config.record_shape
        self.profile_memory = config.profiler.sub_profiler_config.profile_memory
        self.with_stack = config.profiler.sub_profiler_config.with_stack
        self.with_modules = config.profiler.sub_profiler_config.with_modules

    def check(self):
        return self.profiler is not None and self.enable_profile

    def _validate_config(self):
        if len(self.rank_filter) == 0:
            self.rank_filter = [
                0,
            ]
        if self.active_steps <= 0:
            raise ValueError(
                f"[Profiler] active_steps must be positive, got {self.active_steps}"
            )

    def check_finished(self):
        finished = self.is_finished
        if self.is_finished:
            self.is_finished = False
        return finished

    def start(self):
        # start the profiler with static config
        if self.enable_profile:
            if self.global_rank not in self.rank_filter:
                return
            self._validate_config()
            if not self.initialized:
                self.initialized = True
                logger.info(f"[Profiler] init profiler for rank {self.global_rank}")
                self.profiler = torch.profiler.profile(
                    activities=[
                        torch.profiler.ProfilerActivity.CPU,
                        torch.profiler.ProfilerActivity.CUDA,
                    ],
                    schedule=torch.profiler.schedule(
                        wait=self.wait_steps,
                        warmup=self.warmup_steps,
                        active=self.active_steps,
                        repeat=1,
                    ),
                    record_shapes=self.record_shape,
                    with_stack=self.with_stack,
                    with_modules=self.with_modules,
                    profile_memory=self.profile_memory,
                )
                self.thread_pool = futures.ThreadPoolExecutor(max_workers=4)

        if self.check():
            if not self.is_started:
                logger.info(f"[Profiler] start to trace for rank: {self.global_rank}")
                self.profiler.start()
                self.is_started = True
                self.is_finished = False

    def start_dynamic(
        self,
        active_steps: int,
        rank_filter: List[int],
        record_shape: bool,
        profile_memory: bool,
        with_stack: bool,
        with_modules: bool,
    ):
        if self.enable_profile:
            if self.global_rank not in rank_filter:
                return
            self._validate_config()
            if not self.initialized:
                self.initialized = True
                logger.info(f"[Profiler] init profiler for rank {self.global_rank}")
                self.active_steps = active_steps
                self.rank_filter = rank_filter
                self.record_shape = record_shape
                self.profile_memory = profile_memory
                self.with_stack = with_stack
                self.with_modules = with_modules
                self.profiler = torch.profiler.profile(
                    activities=[
                        torch.profiler.ProfilerActivity.CPU,
                        torch.profiler.ProfilerActivity.CUDA,
                    ],
                    schedule=torch.profiler.schedule(
                        wait=self.wait_steps,
                        warmup=self.warmup_steps,
                        active=self.active_steps,
                        repeat=1,
                    ),
                    record_shapes=self.record_shape,
                    with_stack=self.with_stack,
                    with_modules=self.with_modules,
                    profile_memory=self.profile_memory,
                )
                self.thread_pool = futures.ThreadPoolExecutor(max_workers=4)

        if self.check():
            if not self.is_started:
                logger.info(f"[Profiler] start to trace for rank: {self.global_rank}")
                self.profiler.start()
                self.is_started = True
                self.is_finished = False

    def stop(self):
        if self.check():
            if self.is_started:
                self.profiler.stop()
            self.is_started = False

    def step(self):
        if self.check() and self.is_started:
            logger.debug(
                f"[Profiler] Step ahead for rank: {self.global_rank}, step_num: {self.profiler.step_num}"
            )
            self.profiler.step()
            if self.should_stop():
                self.stop_and_save()

    def get_total_steps_needed(self):
        if self.profiler is not None:
            warmup_steps = self.warmup_steps
            wait_steps = self.wait_steps
            active_steps = self.active_steps
            num_steps = wait_steps + warmup_steps + active_steps
            return num_steps
        return 0

    def should_stop(self):
        if self.check():
            self.is_finished = self.profiler.step_num >= self.get_total_steps_needed()
            return self.profiler.step_num >= self.get_total_steps_needed()
        return True

    def stop_and_save(self):
        if self.check():
            self.stop()
            self.save()
            # shutdown the thread pool
            self.thread_pool.shutdown(wait=True)
            # relase the profiler
            del self.profiler
            self.profiler = None
            del self.thread_pool
            self.thread_pool = None
            self.is_started = False
            self.profiled_times += 1
            self.initialized = False
            logger.info(f"[Profiler] stop trace for rank: {self.global_rank}")

    def save(self):
        if self.check():
            os.makedirs(self.output_dir, exist_ok=True)
            trace_file_name = "trace.json.gz"
            trace_file_path = os.path.join(
                self.output_dir, f"{self.profiled_times}_{trace_file_name}"
            )
            logger.info(
                f"[Profiler] save trace for rank: {self.global_rank} to file: {trace_file_path} after {self.profiler.step_num} steps."
            )
            # save the trace asynchronously
            self.profiler.export_chrome_trace(trace_file_path)

            # report to the controller
            # only report the dir
            self.report_to_controller(os.path.dirname(trace_file_path))

    def report_to_controller(self, trace_file_dir: str):
        abs_path = os.path.abspath(trace_file_dir)
        if self.config.train.ckpt.upload_s3:
            # FIXME: (lms) Now we do not have the authority to S3 storage, test it later.
            logger.info(
                f"Uploading the trace to s3: {self.config.train.ckpt.s3_bucket}..."
            )
            self.thread_pool.submit(
                upload_folder_to_s3,
                abs_path,
                self.config.train.ckpt.s3_bucket,
                self.config.train.ckpt.s3_prefix,
            )
        else:
            # just leave the trace file in local disk.
            pass
        post_func = partial(
            requests.post,
            json={
                "replica_name": self.replica_name,
                "trace_path": abs_path,
                "global_rank": self.global_rank,
            },
        )

        report_func = partial(
            make_request_with_retry,
            post_func,
            self.alternative_urls,
            max_retries=constant.COSMOS_HTTP_RETRY_CONFIG.max_retries,
        )
        self.thread_pool.submit(report_func)



import unittest

from torch import nn
from cosmos_rl.utils.fp8.fp8_util import (
    FP8ModelConverter,
    IS_TORCH_COMPATIBLE_WITH_FP8,
)
from cosmos_rl.policy.config import Config as CosmosConfig
from cosmos_rl.policy.config import FP8Config
from cosmos_rl.utils.parallelism import ParallelDims

from torchao.float8.float8_linear import Float8Linear


class DemoModel(nn.Module):
    def __init__(self, input_dim, output_dim, intermediate_dim):
        super().__init__()
        self.norm = nn.LayerNorm(input_dim)
        self.linear = nn.Linear(input_dim, intermediate_dim)
        self.linear2 = nn.Linear(intermediate_dim, output_dim)

    def forward(self, x):
        return self.linear2(self.linear(self.norm(x)))

    def fqn_filter_for_fp8(self):
        return []


@unittest.skipIf(
    not IS_TORCH_COMPATIBLE_WITH_FP8, "FP8 is not supported for this version of PyTorch"
)
class TestFp8(unittest.TestCase):
    def _test_fp8_model_converter_delayed_scaling(self, quant_recipe, dp_shard):
        demo_model = DemoModel(input_dim=1024, output_dim=512, intermediate_dim=1024)
        config = CosmosConfig()
        fp8_config = FP8Config(
            enable_fp8=True, quant_recipe=quant_recipe, fp8_recipe="delayed_scaling"
        )
        config.train.fp8 = fp8_config

        # Mock the parallel_dims
        tp_size = 2
        parallel_dims = ParallelDims(
            dp_replicate=1,
            dp_shard=dp_shard,
            cp=1,
            tp=tp_size,
            pp=1,
            world_size=dp_shard * tp_size,
            pp_dynamic_shape=False,
        )
        converter = FP8ModelConverter(config, parallel_dims)
        converter.convert_model(demo_model)

    def test_fp8_model_converter_delayed_scaling(self):
        params = [
            ("rowwise", 1),
            ("tensorwise", 1),
            ("rowwise", 2),
            ("tensorwise", 2),
        ]
        for quant_recipe, dp_shard in params:
            with self.subTest(quant_recipe=quant_recipe, dp_shard=dp_shard):
                # right now expect NotImplementedError for delayed scaling
                with self.assertRaises(NotImplementedError):
                    self._test_fp8_model_converter_delayed_scaling(
                        quant_recipe=quant_recipe, dp_shard=dp_shard
                    )

    def test_fp8_model_converter(self):
        params = [
            ("rowwise", "dynamic_scaling", 1),
            ("tensorwise", "dynamic_scaling", 1),
            ("rowwise", "dynamic_scaling", 2),
            ("tensorwise", "dynamic_scaling", 2),
        ]
        for quant_recipe, fp8_recipe, dp_shard in params:
            with self.subTest(
                quant_recipe=quant_recipe, fp8_recipe=fp8_recipe, dp_shard=dp_shard
            ):
                self._test_fp8_model_converter(
                    quant_recipe=quant_recipe,
                    fp8_recipe=fp8_recipe,
                    dp_shard=dp_shard,
                )

    def _test_fp8_model_converter(self, quant_recipe, fp8_recipe, dp_shard):
        demo_model = DemoModel(input_dim=1024, output_dim=512, intermediate_dim=1024)
        config = CosmosConfig()
        fp8_config = FP8Config(
            enable_fp8=True, quant_recipe=quant_recipe, fp8_recipe=fp8_recipe
        )
        config.train.fp8 = fp8_config

        # Mock the parallel_dims
        tp_size = 2
        parallel_dims = ParallelDims(
            dp_replicate=1,
            dp_shard=dp_shard,
            cp=1,
            tp=tp_size,
            pp=1,
            world_size=dp_shard * tp_size,
            pp_dynamic_shape=False,
        )
        config.train.fp8.quant_recipe = quant_recipe
        config.train.fp8.fp8_recipe = fp8_recipe

        converter = FP8ModelConverter(config, parallel_dims)
        converter.convert_model(demo_model)

        # Check the model is converted
        assert isinstance(
            demo_model.linear, Float8Linear
        ), f"Got type: {type(demo_model.linear)}"
        assert isinstance(
            demo_model.linear2, Float8Linear
        ), f"Got type: {type(demo_model.linear2)}"


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import torch
import threading
from cosmos_rl.utils.logging import logger
from cosmos_rl.utils.checkpoint import (
    upload_folder_to_s3,
    CheckpointMananger,
)
from transformers import AutoTokenizer, AutoConfig, AutoProcessor, GenerationConfig
from cosmos_rl.policy.trainer.optm import build_optimizers, build_lr_schedulers
from cosmos_rl.policy.model import ModelRegistry
from cosmos_rl.policy.config import Config as CosmosConfig
from cosmos_rl.utils.parallelism import ParallelDims
from cosmos_rl.dispatcher.protocol import Role
from cosmos_rl.comm.base import CommMixin
from safetensors.torch import save_file
from huggingface_hub import create_repo, upload_folder, whoami
from huggingface_hub.utils import disable_progress_bars, enable_progress_bars
from typing import Dict
import cosmos_rl.utils.util as util
from cosmos_rl.utils.profiler import CosmosProfiler
from cosmos_rl.utils.api_suffix import COSMOS_API_SET_TRACE_PATH_SUFFIX
from cosmos_rl.utils.fp8.fp8_util import FP8ModelConverter


class Trainer(CommMixin):
    def __init__(self, config: CosmosConfig, parallel_dims: ParallelDims):
        super().__init__()
        self.config = config
        if self.config.policy.parallelism.dp_shard_size == -1:
            self.config.policy.parallelism.dp_shard_size = parallel_dims.dp_shard
        self.parallel_dims = parallel_dims
        self.local_rank = int(os.environ.get("LOCAL_RANK", 0))
        self.global_rank = int(os.environ.get("RANK", 0))
        self.role = Role.POLICY
        self.world_size = int(os.environ.get("WORLD_SIZE", 1))
        self.device = torch.device(f"cuda:{self.local_rank}")
        torch.cuda.set_device(self.device)
        self.check_config()
        self.tokenizer = util.retry(AutoTokenizer.from_pretrained)(
            config.policy.model_name_or_path,
            trust_remote_code=True,
        )
        # Ensure pad_token_id is set; fallback to eos_token_id if missing (e.g., for models like Mistral)
        if getattr(self.tokenizer, "pad_token_id", None) is None:
            try:
                logger.warning(
                    f"Tokenizer for {config.policy.model_name_or_path} has no pad_token_id, try to use eos_token_id({self.tokenizer.eos_token_id}) as pad_token_id"
                )
                self.tokenizer.pad_token_id = self.tokenizer.eos_token_id
            except Exception as e:
                logger.warning(
                    f"Failed to set pad_token_id with eos_token_id, error = {e}, ignore if not needed"
                )

        self.hf_config = util.retry(AutoConfig.from_pretrained)(
            config.policy.model_name_or_path,
            trust_remote_code=True,
        )
        try:
            self.hf_processor = util.retry(AutoProcessor.from_pretrained)(
                config.policy.model_name_or_path,
                trust_remote_code=True,
            )
        except Exception as e:
            self.hf_processor = None
            logger.info(
                f"Failed to load processor for {config.policy.model_name_or_path}, using tokenizer as processor, ignore if not needed, error = {e}"
            )

        self.train_stream = torch.cuda.current_stream()
        self.init_comm()
        model = ModelRegistry.build_model(config)

        # FP8 settings
        with torch.device("meta"):
            if config.train.fp8.enable_fp8:
                self.model_converter = FP8ModelConverter(config, parallel_dims)
                self.model_converter.convert_model(model)

        if config.train.fsdp_offload:
            model.to_empty(device="cpu")

        try:
            # Apply parallelism to the model
            parallelize_fn, _ = model.parallelize_fn
            # `pp_scheduler` is used for both `sft` and `RLHF`
            # `pp_scheduler_val` is used only for `sft`, since `RLHF` does not require policy model via validation
            self.pp_scheduler, self.pp_scheduler_val = parallelize_fn(
                model, parallel_dims, config, pp_loss_fn=self.pp_loss_fn
            )
            if not config.train.fsdp_offload:
                model.to_empty(device=self.device)
            model.post_to_empty_hook(config)
            torch.cuda.empty_cache()
            self.model_parts = model.separate_model_parts()
            self.model = model
            # util.add_nan_checks(model)
        except Exception as e:
            import traceback

            traceback.print_exc()
            raise e

        self.ckpt_manager = CheckpointMananger(
            config, self.parallel_dims, self.global_rank
        )
        # profiler is initialized after the init_comm()
        self.profiler = CosmosProfiler(
            config,
            parallel_dims,
            replica_name=self.replica_name,
            alternative_urls=self.get_alternative_urls(
                COSMOS_API_SET_TRACE_PATH_SUFFIX
            ),
        )

        self.report_data = {}
        # TODO(cjx): add `CompiledAutograd` support
        self.optimizers = build_optimizers(self.model_parts, self.config)

        if self.config.train.fp8.enable_fp8:
            self.optimizers.register_step_post_hook(
                lambda *args, **kwargs: self.model_converter.post_optimizer_hook(
                    self.model_parts
                )
            )

        self.lr_schedulers = build_lr_schedulers(self.optimizers, self.config)
        self.seq_len_multiple = parallel_dims.cp * parallel_dims.tp
        if self.config.train.fp8.enable_fp8:
            # Constraint of FP8 kernel(torch._scaled_mm): it requires K in MNK is mutiple of 16. In backward of Linear, to
            # calculate the gradient of weight, we have to round the seq_len_multiple to mutiple of 16.
            # See: https://github.com/pytorch/pytorch/blob/851a6fa82df251fbc368f0284d941ce78f68e7b1/aten/src/ATen/native/cuda/Blas.cpp#L1252
            self.seq_len_multiple = (self.seq_len_multiple + 16 - 1) // 16 * 16
            logger.info(
                "FP8 Training enabled, round seq_len_multiple to mutiple of 16."
            )
        logger.info(
            f"Trainer initialized at local rank {self.local_rank}, with seq_len_multiple: {self.seq_len_multiple}"
        )
        self.upload_thread = None

    def check_config(self):
        mini_batch = 1
        policy_type = self.config.train.train_policy.type
        train_batch_per_replica = self.config.train.train_batch_per_replica
        dp_shard_size = self.config.policy.parallelism.dp_shard_size
        error_msg = f"train_batch_per_replica({train_batch_per_replica}) of {policy_type} must be divisible by dp_shard_size({dp_shard_size})"
        if policy_type == "grpo":
            mini_batch = self.config.train.train_policy.mini_batch
            error_msg += f" * mini_batch({mini_batch})"
        assert dp_shard_size == self.parallel_dims.dp_shard
        assert dp_shard_size > 0, "dp_shard_size must be greater than 0"
        assert train_batch_per_replica % (dp_shard_size * mini_batch) == 0, error_msg
        logger.info("Config checked successfully")

    @property
    def pp_loss_fn(self):
        raise NotImplementedError("pp_loss_fn must be provided by subclass")

    def export_safetensors(
        self,
        output_dir: str,
        rel_path: str,
        trainable_only: bool = False,
        is_final=False,
    ):
        path = os.path.join(output_dir, rel_path)
        if self.parallel_dims.dp_replicate_coord[0] > 0:
            return

        if self.global_rank == 0:
            logger.info(
                f"Prepare to exporting safetensors to {path} at rank {self.global_rank}"
            )
        torch.distributed.barrier()

        def get_tensor_size(tensor):
            """Get the size of the tensor in bytes."""
            return tensor.element_size() * tensor.numel()

        max_file_size_gb = 4
        max_size_bytes = max_file_size_gb * 1024**3  # 4 GB in bytes
        current_chunk = {}
        total_chunk_size = 0
        current_chunk_size = 0
        file_idx = 0
        manifest = {}  # Record the weight->file name mapping

        def create_file_name(pp_rank, pp_size, file_idx):
            if pp_size == 1:
                name = f"{file_idx:05d}.safetensors"
            else:
                name = f"model-{pp_rank}-of-{pp_size}-{file_idx:05d}.safetensors"
            return os.path.join(path, name)

        def save_chunked_tensors(
            chunk: Dict[str, torch.Tensor], chunk_size: int, file_path: str
        ):
            """
            Save a dictionary of tensors into a safetensors file.
            Only the rank 0 of dp_shard, cp, tp will save the file.
            """
            nonlocal total_chunk_size
            if (
                self.parallel_dims.dp_shard_coord[0] == 0
                and self.parallel_dims.cp_coord[0] == 0
                and self.parallel_dims.tp_coord[0] == 0
            ):
                os.makedirs(os.path.dirname(file_path), exist_ok=True)
                for name, param in chunk.items():
                    manifest[name] = os.path.basename(file_path)
                total_chunk_size += chunk_size
                save_file(chunk, file_path)
                logger.info(f"Saved chunk {file_idx} to {os.path.basename(file_path)}")

        for name, param in self.model.named_parameters():
            # First map the key from local to hf naming convention
            name = self.model.weight_mapper.policy_map_local_key_to_hf_key(name)
            if trainable_only and not param.requires_grad:
                continue
            is_dtensor = isinstance(param, torch.distributed.tensor.DTensor)
            param = param.full_tensor() if is_dtensor else param
            param = param.detach().data

            pp_rank, pp_size = self.parallel_dims.pp_coord

            for (
                _name,
                _param,
            ) in self.model.weight_mapper.policy_maybe_decompose_weights_to_hf_naming(
                name, param
            ):
                if _param is None:
                    logger.debug(
                        f"[Policy] Skipping None parameter for {name} in safetensors export."
                    )
                    continue
                tensor_size = get_tensor_size(_param)
                # If adding the current tensor exceeds the size limit, save the current chunk
                if current_chunk_size + tensor_size > max_size_bytes:
                    # Save the current chunk as a safetensor file
                    file_name = create_file_name(pp_rank, pp_size, file_idx)
                    save_chunked_tensors(current_chunk, current_chunk_size, file_name)

                    # Reset for the next chunk
                    current_chunk = {_name: _param}
                    current_chunk_size = tensor_size
                    file_idx += 1
                else:
                    # Add tensor to the current chunk
                    current_chunk[_name] = _param
                    current_chunk_size += tensor_size

        # Save any remaining tensors in the last chunk
        if current_chunk:
            file_name = create_file_name(pp_rank, pp_size, file_idx)
            save_chunked_tensors(current_chunk, current_chunk_size, file_name)

        # Allgather the manifest from all pipeline stages
        if self.parallel_dims.pp_enabled:
            pp_group = self.parallel_dims.mesh["pp"].get_group()
            pp_size = self.parallel_dims.pp
            output = [None for _ in range(pp_size)]
            tensor_sizes = [0 for _ in range(pp_size)]
            torch.distributed.all_gather_object(output, manifest, group=pp_group)
            torch.distributed.all_gather_object(
                tensor_sizes, total_chunk_size, group=pp_group
            )
            merged_manifest = {}
            for m in output:
                merged_manifest.update(m)
            total_tensor_size = sum(tensor_sizes)
        else:
            merged_manifest = manifest
            total_tensor_size = total_chunk_size

        torch.distributed.barrier()

        def upload_handler(config, is_final, path, rel_path, max_retries=3):
            """Handle the upload of the model to huggingface and s3."""
            # upload the final model to huggingface
            if config.train.ckpt.upload_hf and is_final:
                username = whoami()["name"]
                repo_id = (
                    username
                    + "/"
                    + config.train.ckpt.hf_repo_name
                    + "-"
                    + config.train.timestamp
                )
                logger.info(f"Uploading the final model to huggingface: {repo_id}...")
                retry = 0
                success = False
                while retry < max_retries:
                    try:
                        create_repo(repo_id, exist_ok=True)
                        # hide redundant logs of huggingface
                        disable_progress_bars()
                        upload_folder(
                            folder_path=path,
                            path_in_repo=".",
                            repo_id=repo_id,
                            commit_message="Upload model",
                        )
                        enable_progress_bars()
                        logger.info(f"Model uploaded to huggingface: {repo_id}")
                        success = True
                        break
                    except Exception as e:
                        logger.error(f"Failed to upload model to huggingface: {e}")
                        retry += 1
                if not success:
                    logger.error(
                        "All retry attempts to upload model to huggingface failed."
                    )
                    raise RuntimeError(
                        f"Failed to upload model to huggingface after {max_retries} attempts."
                    )
            # upload the model to s3
            if config.train.ckpt.upload_s3:
                if is_final:
                    # syncronizely upload the final model to s3
                    upload_folder_to_s3(
                        path,
                        config.train.ckpt.s3_bucket,
                        os.path.join(config.train.ckpt.s3_prefix, rel_path),
                    )
                elif config.train.ckpt.upload_s3 == "all":
                    # asynchronously upload the model to s3
                    upload_folder_to_s3(
                        path,
                        config.train.ckpt.s3_bucket,
                        os.path.join(config.train.ckpt.s3_prefix, rel_path),
                    )
            logger.info(f"\n\nExported safetensors to {path}\n\n")

        if self.global_rank == 0:
            with open(os.path.join(path, "model.safetensors.index.json"), "w") as f:
                json.dump(
                    {
                        "metadata": {
                            "total_size": total_tensor_size,
                        },
                        "weight_map": merged_manifest,
                    },
                    f,
                    indent=4,
                )
            # save hf_config and tokenizer_config
            self.hf_config.save_pretrained(path)
            self.tokenizer.save_pretrained(path)
            if self.hf_processor is not None:
                self.hf_processor.save_pretrained(path)
            # save the generation config to get the generation aligned with HF.
            try:
                generation_config = util.retry(GenerationConfig.from_pretrained)(
                    self.config.policy.model_name_or_path
                )
                generation_config.save_pretrained(path)
            except Exception:
                logger.warning("[Policy] No generation config found, do not save it.")

            need_upload = (
                self.config.train.ckpt.upload_hf and is_final
            ) or self.config.train.ckpt.upload_s3
            if need_upload:
                # If the upload thread is already running, wait for it to finish
                if self.upload_thread is not None:
                    self.upload_thread.join()
                self.upload_thread = threading.Thread(
                    target=upload_handler,
                    args=(self.config, is_final, path, rel_path),
                    name="upload_safetensors",
                    daemon=True,
                )
                self.upload_thread.start()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import msgpack
import torch
from cosmos_rl.utils.parallelism import ParallelDims
from typing import Dict, List, Tuple, Callable, Any, Optional, Union
from cosmos_rl.utils.constant import COSMOS_HF_MODEL_TYPES
from cosmos_rl.policy.model.base import WeightMapper
from cosmos_rl.utils.logging import logger
from vllm.model_executor.layers.linear import (
    RowParallelLinear,
    ColumnParallelLinear,
    QKVParallelLinear,
    MergedColumnParallelLinear,
)
from vllm.model_executor.layers.vocab_parallel_embedding import VocabParallelEmbedding

from torch.nn.parameter import Parameter
from math import gcd
from functools import reduce
import asyncio
from cosmos_rl.utils import util
import multiprocessing
from concurrent.futures import ProcessPoolExecutor
from cosmos_rl.policy.config import Config as CosmosConfig


class DimSliceInfo:
    """
    A class to represent the slice information of a tensor along a specific dimension.
    This class contains the offset, total size, dimension name, and length of the slice.
    """

    offset: int
    total_size: int
    dim: str
    length: int = 1

    def __init__(self, offset: int, total_size: int, dim: str = "", length: int = 1):
        """
        Initialize the DimSliceInfo with the given offset, total size, dimension name, and length.
        """
        self.offset = offset
        self.total_size = total_size
        self.dim = dim
        self.length = length

    def __repr__(self):
        # Returning a dictionary representation
        return f"{self.__dict__}"

    @classmethod
    def from_dict(cls, data: Dict[str, Any]):
        """
        Create a DimSliceInfo object from a dictionary.
        :param data: A dictionary containing the keys 'offset', 'total_size', 'dim', and 'length'.
        :return: A DimSliceInfo object.
        """
        return DimSliceInfo(
            offset=data["offset"],
            total_size=data["total_size"],
            dim=data.get("dim", ""),
            length=data.get("length", 1),
        )

    def simplify(self):
        common = reduce(gcd, [self.offset, self.total_size, self.length])  # noqa: E741
        return DimSliceInfo(
            offset=self.offset // common,
            total_size=self.total_size // common,
            dim=self.dim,
            length=self.length // common,
        )


def slice_tensor_with_strategy(
    tensor: torch.Tensor, idx: int, tensor_split_strategy: DimSliceInfo
):
    """
    Slices a tensor according to the given strategy at one dimension index.
    :param tensor: The tensor to be sliced.
    :param idx: The index of the dimension to slice.
    :param tensor_split_strategy: The strategy for slicing the tensor.
    :return: A sliced view of the tensor for the given dimension index.
    """

    view = tensor
    assert view.shape[idx] % tensor_split_strategy.total_size == 0
    start = (
        view.shape[idx]
        // tensor_split_strategy.total_size
        * tensor_split_strategy.offset
    )
    length = (
        view.shape[idx]
        // tensor_split_strategy.total_size
        * tensor_split_strategy.length
    )
    dim = view.dim()
    assert idx < view.dim(), f"Invalid index {idx} for {dim}D tensor."
    slices = (
        [slice(None, None)] * idx
        + [slice(start, start + length)]
        + [slice(None, None)] * (dim - idx - 1)
    )
    return view[slices]


def slice_tensor_with_strategies(
    self: torch.Tensor, strategys: Dict[int, Union[DimSliceInfo, Dict[str, Any]]]
) -> torch.Tensor:
    """
    Slices the tensor according to the given strategies at all dimension indices.
    :param tensor: The tensor to be sliced.
    :param strategys: A dictionary mapping dimension indices to DimSliceInfo objects.
    :return: The sliced tensor.
    """
    view = self
    for idx, split in strategys.items():
        idx = int(idx)
        if isinstance(split, dict):
            split = DimSliceInfo.from_dict(split)
        view = slice_tensor_with_strategy(view, idx, split)
    return view


torch.Tensor.cosmos_slice = slice_tensor_with_strategies


class WeightSyncInstruction:
    """
    A class to represent a weight synchronization instruction for a specific parameter.
    This class contains the parameter name, the global rank, and the instruction.
    """

    def __init__(
        self,
        policy_rank: int,
        rollout_rank: int,
        slice_strategy: Dict[int, DimSliceInfo],
    ):
        """
        Initialize the WeightSyncInstruction with the given policy rank, rollout rank, and slice strategy.
        :param policy_rank: The rank of the policy in the parallelism configuration.
        :param rollout_rank: The rank of the rollout in the parallelism configuration.
        :param slice_strategy: A dictionary mapping dimension indices to DimSliceInfo objects representing the slicing strategy.
        """
        self.policy_rank = policy_rank
        self.rollout_rank = rollout_rank
        self.slice_strategy = slice_strategy

    def __repr__(self):
        # Returning a dictionary representation
        return f"{self.__dict__}"


class WeightSyncInstructionsPerParam:
    """
    A class to represent a collection of weight synchronization instructions for a specific param.
    This class contains the parameter name and a list of synchronization instructions.
    """

    def __init__(self, param_name: str, instructions: List[WeightSyncInstruction]):
        """
        Initialize the WeightSyncInstructionsPerParam with the given parameter name and instructions.
        :param param_name: The name of the parameter for which the instructions are created.
        :param instructions: A list of WeightSyncInstruction objects representing the synchronization instructions.
        """
        self.param_name = param_name
        self.instructions = instructions

    def __repr__(self):
        # Returning a dictionary representation
        return f"{self.__dict__}"


class WeightSyncInstructionsGroup:
    """
    A class to represent a group of weight synchronization instructions for multiple parameters.
    This class contains a list of WeightSyncInstructionsPerParam objects.
    """

    def __init__(self, param_instructions: List[WeightSyncInstructionsPerParam]):
        """
        Initialize the WeightSyncInstructionsGroup with the given instructions.
        :param param_instructions: A list of WeightSyncInstructionsPerParam objects representing the synchronization instructions for multiple parameters in one group.
        """
        self.param_instructions = param_instructions

    def __repr__(self):
        # Returning a dictionary representation
        return f"{self.__dict__}"

    @classmethod
    def from_dict(cls, data: Dict[str, Any]):
        """
        Create a WeightSyncInstructionsGroup object from a dictionary.
        :param data: A dictionary containing the keys 'instructions' and 'param_names'.
        :return: A WeightSyncInstructionsGroup object.
        """
        instructions = [
            WeightSyncInstructionsPerParam(
                param_name=insts["param_name"],
                instructions=[
                    WeightSyncInstruction(
                        policy_rank=inst["policy_rank"],
                        rollout_rank=inst["rollout_rank"],
                        slice_strategy={
                            k: DimSliceInfo.from_dict(v)
                            for k, v in inst["slice_strategy"].items()
                        },
                    )
                    for inst in insts["instructions"]
                ],
            )
            for insts in data["param_instructions"]
        ]
        return cls(instructions)


class ParallelTopoMapper:
    """
    A class used for weight sharing topology map for weight synchronization.
    """

    ordered_dims: List[str] = ["tp", "dp_shard_cp", "dp_cp_tp"]

    def __init__(
        self,
        parallelism: Optional[ParallelDims],
        parallelism_strategy: Optional[Callable],
        weight_mapper: WeightMapper,
        hf_config: Any,
        is_policy: bool,
        underlying_model: Any,
    ):
        """
        Initialize the ParallelTopoMap with the given parallelism configurations.

        :param parallelism: The parallelism config for the policy or rollout.
        :param parallelism_strategy: The strategy function for the policy parallelism or rollout parallelism if specified.
        :param weight_mapper: The weight mapper to use for mapping weights.
        :param hf_config: The huggingface config.
        :param is_policy: A boolean indicating if this is for policy or rollout.
        :param underlying_model: The underlying model for which the parallelism map is created.
        """
        self.parallelism = parallelism
        self.parallelism_strategy = parallelism_strategy
        ranks = range(self.parallelism.world_size)
        full_mesh_rank_info_map = []
        for r in ranks:
            full_rank = self.get_full_mesh_rank_info(r)
            full_mesh_rank_info_map.append(full_rank)
        self.full_mesh_rank_info_map = full_mesh_rank_info_map
        self.hf_config = hf_config
        self.weight_mapper = weight_mapper
        self.is_policy = is_policy
        self.underlying_model = underlying_model

    def get_full_mesh_rank_info(self, global_rank: int) -> Dict[str, DimSliceInfo]:
        """
        Get the full mesh rank info of the given global rank in the simulation map.

        :param global_rank: The global rank to get the full rank for.
        :return: A dictionary mapping each parallel mesh dimension to its mesh rank info represented using DimSliceInfo.
        """
        full_mesh_rank_info = {}
        for dim in self.ordered_dims:
            full_mesh_rank_info[dim] = DimSliceInfo(
                self.parallelism.get_rank_in_dim(dim, global_rank),
                self.parallelism.get_size_in_dim(dim),
                dim,
            )
        full_mesh_rank_info["pp"] = DimSliceInfo(
            self.parallelism.get_rank_in_dim("pp", global_rank),
            self.parallelism.get_size_in_dim("pp"),
            "pp",
        )
        return full_mesh_rank_info

    @classmethod
    def get_unified_rank_info(
        cls, a: DimSliceInfo, b: DimSliceInfo
    ) -> Tuple[DimSliceInfo, DimSliceInfo]:
        """
        Get the unified slice information with the same total size for two DimSliceInfo objects.
        :param a: The first DimSliceInfo object.
        :param b: The second DimSliceInfo object.
        :return: A tuple containing the unified slice information for both objects.
        """
        size = max(a.total_size, b.total_size)
        assert (
            size % a.total_size == 0 and size % b.total_size == 0
        ), "Sizes are not compatible for unification"
        scale_a = size // a.total_size
        scale_b = size // b.total_size
        scaled_a_size = a.total_size * scale_a
        scaled_b_size = b.total_size * scale_b
        scaled_a_rank = a.offset * scale_a
        scaled_b_rank = b.offset * scale_b
        unified_a = DimSliceInfo(
            scaled_a_rank, scaled_a_size, a.dim, a.length * scale_a
        )
        unified_b = DimSliceInfo(
            scaled_b_rank, scaled_b_size, b.dim, b.length * scale_b
        )
        return unified_a, unified_b

    @classmethod
    def rank_overlap(cls, a: DimSliceInfo, b: DimSliceInfo) -> DimSliceInfo:
        """
        Check if the parts of two DimSliceInfo objects overlap.

        :param a: The first DimSliceInfo object.
        :param b: The second DimSliceInfo object.
        :return: A DimSliceInfo object representing the overlap, or None if there is no overlap.
        """
        a_new, b_new = cls.get_unified_rank_info(a, b)
        assert (
            a_new.total_size == b_new.total_size
        ), "Sizes do not match after unification"

        left = max(a_new.offset, b_new.offset)
        right = min(
            a_new.offset + a_new.length,
            b_new.offset + b_new.length,
        )
        overlapped = None
        if left < right:
            overlapped = DimSliceInfo(left, a_new.total_size, a_new.dim, right - left)
        return overlapped

    @classmethod
    def relative_rank(cls, smaller: DimSliceInfo, larger: DimSliceInfo) -> DimSliceInfo:
        """
        Get the relative slice information of two DimSliceInfo objects.
        :param smaller: The smaller DimSliceInfo object.
        :param larger: The larger DimSliceInfo object.
        :return: A DimSliceInfo object representing the relative slice of the smaller on in the larger one.
        """
        s, l = cls.get_unified_rank_info(smaller, larger)  # noqa: E741
        assert (
            s.offset >= l.offset
        ), "Smaller rank is not less than or equal to larger rank"
        assert (
            s.offset + s.length <= l.offset + l.length
        ), "Smaller rank does not fit within larger rank"
        rank = s.offset - l.offset
        size = l.length
        length = s.length
        return DimSliceInfo(rank, size, s.dim, length)

    @classmethod
    def merge_rank(cls, outter: DimSliceInfo, inner: DimSliceInfo) -> DimSliceInfo:
        """
        Merge two nested DimSliceInfo objects into one.
        :param outter: The DimSliceInfo object at a outter dimension.
        :param inner: The DimSliceInfo object at an inner dimension.
        :return: A DimSliceInfo object representing the merged slice information.
        """
        assert outter.length == 1, "Outer rank length must be 1"
        size = outter.total_size * inner.total_size
        rank = outter.offset * inner.total_size + inner.offset
        length = inner.length
        return DimSliceInfo(rank, size, outter.dim, length)

    @classmethod
    def tensor_overlap_info_at_dim(
        cls,
        policy_rank: Dict[int, DimSliceInfo],
        rollout_rank: Dict[int, DimSliceInfo],
        dim: int,
    ) -> Tuple[DimSliceInfo, DimSliceInfo]:
        """
        Get the tensor overlap information at one dimension index.
        :param policy_rank: The sharded slice information for the given tensor from policy.
        :param rollout_rank: The sharded slice information for the given tensor from rollout.
        :param dim: The dimension index to check for overlap.
        :return: A tuple containing the overlap information between the given policy and rollout tensors.
        """
        if dim not in policy_rank:
            p = DimSliceInfo(0, 1)
        else:
            p = policy_rank[dim]
        if dim not in rollout_rank:
            r = DimSliceInfo(0, 1)
        else:
            r = rollout_rank[dim]

        p_new, r_new = cls.get_unified_rank_info(p, r)
        overlap = cls.rank_overlap(p_new, r_new)

        if overlap is None:
            return None, None
        overlap_r = cls.relative_rank(overlap, r_new)
        overlap_p = cls.relative_rank(overlap, p_new)
        return overlap_p.simplify(), overlap_r.simplify()

    def shard_info_at_dim(
        self,
        rank_infos: Dict[str, DimSliceInfo],
        dim: str,
    ) -> DimSliceInfo:
        """
        Get the sharded slice information at one mesh dimension.
        :param rank_infos: The shared slice information for all related mesh dimensions.
        :param dim: The dimension to get the shard information for.
        :return: A DimSliceInfo object representing the sharded slice information for the given dimension.
        """
        if dim not in rank_infos:
            p = DimSliceInfo(0, 1, dim)
        else:
            p = rank_infos[dim]

        return p

    def merged_shard_info_at_dim(
        self,
        rank_info: Dict[str, DimSliceInfo],
    ) -> DimSliceInfo:
        """
        Get the merged sharded slice information for different mesh dimensions.
        :param rank_info: The slice information for the mesh dimensions.
        :return: A DimSliceInfo object representing the merged sharded slice information for the given dimensions.
        """

        if self.ordered_dims[0] not in rank_info:
            rank_info[self.ordered_dims[0]] = DimSliceInfo(0, 1, self.ordered_dims[0])
        if self.ordered_dims[1] not in rank_info:
            rank_info[self.ordered_dims[1]] = DimSliceInfo(0, 1, self.ordered_dims[1])
        # Merge the ranks of the two dimensions
        p = self.merge_rank(
            rank_info[self.ordered_dims[0]], rank_info[self.ordered_dims[1]]
        )
        return p

    @classmethod
    def get_global_ranks_for_given_mesh_rank(
        cls, parallel_dims: ParallelDims, mesh_rank: Dict[str, int]
    ) -> List[int]:
        """
        Get the global ranks for a given mesh rank in the parallelism configuration.
        mesh_rank is subset of parallel_dims, so there could be multiple devices have
        the same mesh_rank.
        :param parallel_dims: The parallelism configuration.
        :param mesh_rank: The mesh rank to get the global ranks whose mesh rank matches the given mesh_rank.
        :return: A list of global ranks.
        """
        if len(mesh_rank) == 0:
            return list(range(parallel_dims.world_size))
        global_ranks = []
        for rank in range(parallel_dims.world_size):
            if all(
                [
                    parallel_dims.get_rank_in_dim(dim, rank) == dimr
                    for dim, dimr in mesh_rank.items()
                ]
            ):
                global_ranks.append(rank)
        return global_ranks

    def duplicate_ranks_at_given_dimensions(
        self, dims: List[str], global_rank: int
    ) -> List[int]:
        """
        Get the duplicate global ranks with same mesh rank info with the given global rank at the specified dimensions.
        :param dims: The dimensions to check for duplicate ranks.
        :param global_rank: The global rank to check.
        :return: A list of duplicate global ranks.
        """
        dims_map = {}
        for dim in dims:
            dims_map[dim] = self.parallelism.get_rank_in_dim(dim, global_rank)

        return ParallelTopoMapper.get_global_ranks_for_given_mesh_rank(
            self.parallelism, dims_map
        )

    @classmethod
    def policy_to_rollout_assign(
        cls, policys: List[int], rollouts: List[int], p_rank: int, r_rank: int
    ) -> Tuple[List[int], List[int]]:
        """
        Assign policy ranks to rollout ranks sharing the same sharded part related between the given policy and rollout rank.
        :param policys: The list of policy ranks sharing the same sharded part.
        :param rollouts: The list of rollout ranks sharing the same sharded part.
        :param p_rank: The given parallelism rank of the policy to consider.
        :param r_rank: The given parallelism rank of the rollout to consider.
        :return: A tuple containing two list: policy assignment and rollout assignment related between the given policy and rollout rank.
        """
        p_assignment = []
        r_assignment = []
        if len(policys) >= len(rollouts):
            for i, p in enumerate(policys):
                if i >= len(rollouts):
                    break
                if p == p_rank and rollouts[i] == r_rank:
                    p_assignment = [r_rank]
                    r_assignment = [p_rank]
        else:
            group_size = ((len(rollouts) - 1) // len(policys)) + 1
            for i, p in enumerate(policys):
                if p == p_rank:
                    rs = rollouts[
                        i * group_size : min(i * group_size + group_size, len(rollouts))
                    ]
                    if r_rank in rs:
                        p_assignment = [r_rank]
                        r_assignment = [p_rank]
        return p_assignment, r_assignment

    def generate_local_shard_info(
        self,
        dim_to_parallel: Dict[int, list[str]],
        rank_info: Dict[str, DimSliceInfo],
    ) -> Dict[int, Dict]:
        """
        Generate detailed shard info for the given dimensions and ranks.
        :param dim_to_parallel: A dictionary mapping dimension indices to parallel mesh dimensions.
        :param rank_info: The shard information of the parallel mesh dimensions.
        :return: A dictionary mapping dimension indices to DimSliceInfo objects to represent the detailed shard strategy.
        """
        shard_dim_info = {}
        for idx, dims in dim_to_parallel.items():
            if len(dims) == 1:
                shard_dim_info[idx] = self.shard_info_at_dim(
                    rank_info, dims[0]
                ).__dict__
            elif len(dims) == 2:
                assert (
                    self.ordered_dims[2] not in dims
                ), f"Invalid dimension mapping: {dims} in generate_slice_strategies for merge"
                assert (
                    self.ordered_dims[0] in dims and self.ordered_dims[1] in dims
                ), f"Invalid dimension mapping: {dims} in generate_slice_strategies for merge"
                shard_dim_info[idx] = self.merged_shard_info_at_dim(rank_info).__dict__
            else:
                raise ValueError(
                    f"Invalid dimension mapping: {dims} in generate_slice_strategies"
                )
        return shard_dim_info

    def local_shard_info_for_params(
        self,
        params: List[Tuple[str, int]],
        global_rank: int,
    ) -> Dict[str, Any]:
        """
        Generate local shard information for the given parameters.
        :param params: The parameters to generate local shard information for.
        :param global_rank: The global rank to generate local shard information for.
        :return: A dictionary containing the generated local shard information.
        """
        local_shard_info_all_params = {}
        if self.is_policy:
            self.parallelism_info_for_dtensor_params()
        else:
            self.parallelism_info_for_vllm_params()

        for dest_name, shape in params:
            split_dim_map, dim_to_parallel, pp_rank, dims_rank_info = (
                None,
                None,
                None,
                None,
            )
            if self.parallelism_strategy is not None:
                # If custom parallelism strategy is provided, use it to get the split dimensions and parallel mapping.
                # The custom parallelism strategy has high priority.
                # The custom parallelism strategy is specified from `get_policy_parallelism_strategy` and `get_rollout_parallelism_strategy` in weight mapper.
                split_dim_map, dim_to_parallel, pp_rank = self.parallelism_strategy(
                    shape, dest_name, self.parallelism, self.hf_config
                )

            if split_dim_map is None and dim_to_parallel is None and pp_rank is None:
                # If no custom parallelism strategy is provided, use the default automatically inferred parallelism info for the parameter.
                split_dim_map, dim_to_parallel, pp_rank, dims_rank_info = (
                    self.parallelism_info_for_param(dest_name)
                )

            if split_dim_map is None and dim_to_parallel is None and pp_rank is None:
                continue

            ranks = self.full_mesh_rank_info_map[global_rank]
            if ranks["pp"].offset != pp_rank:
                continue

            local_shard_info_all_params[dest_name] = (
                self.generate_local_shard_info(dim_to_parallel, ranks)
                if dims_rank_info is None
                else dims_rank_info
            )
        # Return a dictionary containing the local shard info for each parameter
        return local_shard_info_all_params

    def parallelism_info_for_param(
        self,
        param_name: str,
    ):
        """
        Get the parallelism info for a specific parameter.
        This method returns a tuple of four elements:
            - dims_map: Dict[str, int]: A mapping of mesh names to corresponding dimension index of the param.
            - tensor_dim_to_parallel_map: Dict[int, List[str]]: A mapping of tensor dimension index to their related parallel mesh dimensions.
            - p_rank: int: The pipeline rank of the parameter.
            - dim_slice_info: Optional[Dict[int, Dict]]: A dictionary containing information for the specific slice of the parameter at each dimension index.
        If the parameter name is not found, it returns None for all elements.
        Args:
            param_name (str): The name of the parameter for which to retrieve parallelism info.
        Returns:
            Tuple[Dict[str, int], Dict[int, List[str]], int, Optional[Dict[int, Dict]]]:
            A tuple containing the dimensions map, tensor dimension to parallel map, pipeline rank, and optional dimension slice info.
        """
        if hasattr(self, "parallelism_info_for_params"):
            if param_name in self.parallelism_info_for_params:
                return self.parallelism_info_for_params[param_name]
        else:
            logger.error("No parallelism info found for the given parameter name.")
        return None, None, None, None

    def insert_to_parallelism_info(
        self,
        param_name: str,
        dims_map: Dict[str, int],
        name_to_hf: Callable,
        packed_modules_mapping: Dict[str, Any] = {},
        dims_rank_info: Optional[Dict[int, Dict]] = None,
    ):
        """
        Insert the parallelism info for the policy model parameters.
        This method updates the `policy_parallelism_info_for_params` dictionary with the parameter name,
        its dimensions map, tensor dimension to parallel map, pipeline rank, and optional dimension slice info.
        Args:
            param_name (str): The name of the parameter.
            dims_map (Dict[str, int]): The dimensions map for the parameter.
            name_to_hf (Callable): A function to map parameter names to Hugging Face names.
            packed_modules_mapping (Dict[str, Any], optional): A mapping of packed module names.
            dims_rank_info (Optional[Dict[int, Dict]], optional): A dictionary containing information for the specific slice of the parameter at each dimension index.
        """
        tensor_dim_to_parallel_map = {}
        for k, v in dims_map.items():
            if v not in tensor_dim_to_parallel_map:
                tensor_dim_to_parallel_map[v] = []
            tensor_dim_to_parallel_map[v].append(k)
        p_rank = self.parallelism.pp_coord[0]
        for k, v in packed_modules_mapping.items():
            assert (
                dims_rank_info is None
            ), f"Packed modules mapping {packed_modules_mapping} should not be used with dims_rank_info {dims_rank_info}."
            if k in param_name:
                for rename in v:
                    name = name_to_hf(param_name).replace(k, rename)
                    self.parallelism_info_for_params[name] = (
                        dims_map,
                        tensor_dim_to_parallel_map,
                        p_rank,
                        dims_rank_info,
                    )
                return
        name = name_to_hf(param_name)
        self.parallelism_info_for_params[name] = (
            dims_map,
            tensor_dim_to_parallel_map,
            p_rank,
            dims_rank_info,
        )

    def parallelism_info_for_dtensor_params(self) -> None:
        """
        Get the parallelism info for the dtensor model parameters.
        Normally, this is used for policy model parameters.
        It analyzes the model's named parameters and extracts the parallel dimensions and their shard information.
        The method checks if the model parameters are distributed tensors (DTensor) and extracts their detailed shard information from DTensor specifications.
        This method updates a dictionary with parameter names as keys and their parallel dimensions with shard information as values.
        """
        assert (
            self.is_policy
        ), "parallelism_info_for_dtensor_params should only be called for policy model."
        if hasattr(self, "parallelism_info_for_params"):
            return self.parallelism_info_for_params
        self.parallelism_info_for_params = {}
        for name, param in self.underlying_model.named_parameters():
            is_dist_tensor = isinstance(param, torch.distributed.tensor.DTensor)
            dims_rank_info = {}
            if not is_dist_tensor:
                dims_map = {}
                global_shape = tuple(param.shape)
            else:
                dims_map = {}
                global_shape = tuple(param.shape)
                mesh = param.device_mesh
                placements = param.placements
                assert (
                    len(placements) == len(mesh.mesh_dim_names)
                ), f"Number of placements {placements} does not match number of mesh dimensions {mesh}."
                for dim, placement in zip(mesh.mesh_dim_names, placements):
                    if placement.is_shard():
                        dims_map[dim] = placement.dim
                    elif placement.is_replicate():
                        pass
                    else:
                        raise ValueError(f"Unsupported placement type: {placement}")
                chunk_meta_list = param.__create_chunk_list__()
                local = param.to_local()
                assert (
                    len(chunk_meta_list) == 1
                ), f"Expected only one chunk meta, but got {len(chunk_meta_list)} for {name}."
                meta = chunk_meta_list[0]
                assert (
                    len(meta.offsets)
                    == len(meta.sizes)
                    == len(global_shape)
                    == len(tuple(local.shape))
                ), f"Offsets {meta.offsets} and sizes {meta.sizes} must match global shape {global_shape} and local shape {tuple(local.shape)}."

                for idx, g_size in enumerate(global_shape):
                    offset = int(meta.offsets[idx])
                    total_size = int(g_size)
                    length = int(meta.sizes[idx])
                    if total_size == length:
                        assert (
                            offset == 0
                        ), f"Expected rank 0 for full size dimension {idx}, but got {offset}."
                    else:
                        dims_rank_info[idx] = DimSliceInfo(
                            offset=offset,
                            total_size=total_size,
                            length=length,
                        ).__dict__
                decomposed_key_and_slices = (
                    self.weight_mapper.policy_decompose_param_1_to_n_for_sync(
                        self.weight_mapper.policy_map_local_key_to_hf_key(name)
                    )
                )
                if decomposed_key_and_slices:
                    for part_name, part_slice in decomposed_key_and_slices:
                        splitted_dim_rank_info = {}
                        part_in_local = {}
                        part_slice = {
                            len(global_shape) + k if k < 0 else k: v
                            for k, v in part_slice.items()
                        }
                        all_dims = part_slice.keys() | dims_rank_info.keys()
                        for dim in all_dims:
                            if dim not in part_slice:
                                dim_slice = DimSliceInfo(
                                    offset=0,
                                    total_size=1,
                                )
                            else:
                                dim_slice = DimSliceInfo.from_dict(part_slice[dim])
                            if dim not in dims_rank_info:
                                assert (
                                    len(global_shape) > dim
                                ), f"Dimension {dim} is out of bounds for global shape {global_shape}."
                                local_part = DimSliceInfo(offset=0, total_size=1)
                            else:
                                local_part = DimSliceInfo.from_dict(dims_rank_info[dim])
                            slice_in_splited, overlap_in_local = (
                                self.tensor_overlap_info_at_dim(
                                    {dim: dim_slice}, {dim: local_part}, dim
                                )
                            )
                            if slice_in_splited is None:
                                splitted_dim_rank_info = None
                                break

                            splitted_dim_rank_info[dim] = slice_in_splited.__dict__
                            part_in_local[dim] = overlap_in_local
                        if splitted_dim_rank_info is not None:
                            self.insert_to_parallelism_info(
                                part_name,
                                dims_map,
                                self.weight_mapper.policy_map_local_key_to_hf_key,
                                dims_rank_info=splitted_dim_rank_info,
                            )
                    continue
            self.insert_to_parallelism_info(
                name,
                dims_map,
                self.weight_mapper.policy_map_local_key_to_hf_key,
                dims_rank_info=dims_rank_info,
            )

    def parallelism_info_for_vllm_params(self):
        """
        Get the parallelism info for the vllm rollout model parameters by analyzing the model's named parameters with vllm instance check.
        Normally, this is used for rollout model parameters.
        It extracts the parallel dimensions and their shard information from the vllm model parameters.
        The method checks if the model parameters are instances of vllm parallel layers (like QKVParallelLinear, MergedColumnParallelLinear, etc.)
        and extracts their detailed shard information from the parameters.
        This method updates a dictionary with parameter names as keys and their parallel dimensions with shard information as values.
        """
        assert (
            not self.is_policy
        ), "parallelism_info_for_vllm_params should only be called for rollout model."
        if hasattr(self, "parallelism_info_for_params"):
            return self.parallelism_info_for_params
        self.parallelism_info_for_params = {}

        for param_name, param in self.underlying_model.named_parameters():
            name_parts = param_name.split(".")
            part = self.underlying_model
            is_bias = False
            should_skip = False
            for part_name in name_parts:
                if hasattr(part, part_name):
                    if isinstance(getattr(part, part_name), Parameter):
                        if part_name == "bias":
                            is_bias = True
                        elif part_name == "weight":
                            is_bias = False
                        else:
                            logger.warning(
                                f"Part {part_name} is not a Parameter. Skipping."
                            )
                            should_skip = True
                        break
                    part = getattr(part, part_name)
                elif str.isdigit(part_name):
                    part = part[int(part_name)]
                else:
                    raise ValueError(f"Part {part_name} not found in {part}. Skipping.")
            if should_skip:
                continue
            dims_map = {}
            if isinstance(part, (QKVParallelLinear)):
                output_dim = getattr(param, "output_dim", 0)
                dims_map["tp"] = output_dim
                assert any(
                    [
                        k in param_name
                        for k in self.weight_mapper.packed_modules_mapping.keys()
                    ]
                ), f"QKVParallelLinear {param_name} is not in packed_modules_mapping {self.weight_mapper.packed_modules_mapping}."
            elif isinstance(part, (MergedColumnParallelLinear)):
                output_dim = getattr(param, "output_dim", 0)
                dims_map["tp"] = output_dim
                assert any(
                    [
                        k in param_name
                        for k in self.weight_mapper.packed_modules_mapping.keys()
                    ]
                ), f"MergedColumnParallelLinear {param_name} is not in packed_modules_mapping {self.weight_mapper.packed_modules_mapping}."
            elif isinstance(part, (RowParallelLinear)):
                input_dim = getattr(param, "input_dim", 1)
                if not is_bias:
                    assert (
                        input_dim is not None
                    ), f"RowParallelLinear {param_name} has no input_dim attribute."
                    dims_map["tp"] = input_dim
            elif isinstance(part, (ColumnParallelLinear)):
                output_dim = getattr(param, "output_dim", 0)
                dims_map["tp"] = output_dim
            elif isinstance(part, VocabParallelEmbedding):
                output_dim = getattr(param, "output_dim", 0)
                assert (
                    not is_bias
                ), f"VocabParallelEmbedding {param_name} should not have bias."
                dims_map["tp"] = output_dim
            else:
                assert (
                    "Parallel" not in part.__class__.__name__
                ), f"Part {part.__class__.__name__} is not a parallel layer. Skipping."

            self.insert_to_parallelism_info(
                param_name,
                dims_map,
                self.weight_mapper._rollout_vllm_name_to_hf,
                self.weight_mapper.packed_modules_mapping,
            )


class ParallelTopoMapperGroup:
    """
    A class to represent a group of weight sharing topology maps used for weight synchronization.
    This class manages multiple ParallelTopoMapper instances, each corresponding to a different parallelism strategy.
    Different model parts may have different parallelism strategies in one whole model.
    It is used to prepare local shard information for parameters based on the parallelism configuration.
    It clusters parameters by model part and prepares local shard information for each part.
    """

    def __init__(
        self,
        global_parallelism: ParallelDims,
        hf_config: Any,
        is_policy: bool,
        underlying_model: Any,
        weight_mapper: Optional[WeightMapper] = None,
    ):
        """
        Initialize the ParallelTopoMapperGroup with the given parallelism configurations.

        :param global_parallelism: The parallelism config for the policy or rollout.
        :param hf_config: The huggingface config.
        :param is_policy: A boolean indicating if this is for policy parallelism.
        :param underlying_model: The underlying model for which the parallelism map is created.
        :param weight_mapper: An optional WeightMapper instance. If None, a default mapper is used based on the model type from hf_config.
        """
        self.hf_config = hf_config
        model_type = hf_config.model_type
        self.mapper_group: List[ParallelTopoMapper] = []

        if weight_mapper is None:
            if model_type not in WeightMapper._MODEL_WEIGHT_MAPPER_REGISTRY:
                logger.warning(
                    f"[ParallelTopoMapperGroup] can not find {model_type} in weight mapper, use {COSMOS_HF_MODEL_TYPES} model type instead."
                )
                model_type = COSMOS_HF_MODEL_TYPES
            weight_mapper_fn = WeightMapper.get_weight_mapper(model_type)
            self.weight_mapper = weight_mapper_fn(hf_config)
        else:
            self.weight_mapper = weight_mapper

        # Note: policy_strategies and rollout_strategies callable to decide if or how to parallel
        # the param tensor of a give name if given as a counterpart of the automatic inferred parallelism strategy.

        # The replica has its single global parallelism config
        # But there may be different parallelism strategies executed by different model parts
        # For example: VLM has 2 parts: the visual encoder and the language encoder.
        #   the visual encoder may merge TP and DP meshes, while the language encoder may not
        # So we may need a ParallelTopoMapper for each model part with its own parallelism strategy
        # So the stategies might be a list containing multiple strategies for different model parts.

        if is_policy:
            strategies = self.weight_mapper.get_policy_parallelism_strategy()
        else:
            strategies = self.weight_mapper.get_rollout_parallelism_strategy()

        if strategies:
            for strategy in strategies:
                self.mapper_group.append(
                    ParallelTopoMapper(
                        global_parallelism,
                        strategy,
                        weight_mapper,
                        hf_config,
                        is_policy=is_policy,
                        underlying_model=underlying_model,
                    )
                )
        else:
            # If no parallelism strategy is provided, reserve it as None to use the default automatic parallelism infer.
            self.mapper_group.append(
                ParallelTopoMapper(
                    global_parallelism,
                    None,
                    weight_mapper,
                    hf_config,
                    is_policy=is_policy,
                    underlying_model=underlying_model,
                )
            )

    def _cluster_params_by_model_part(
        self, params: List[Tuple[str, int]]
    ) -> List[List[Tuple[str, int]]]:
        """
        Resort the parameters based on the name mapper.
        :param params: The parameters to resort.
        :return: A list of tuples containing the resorted parameters separated by different model parts.
        """
        if len(self.mapper_group) == 1:
            return [params]
        x = [[] for _ in self.mapper_group]
        for name, rank in params:
            idx = self.weight_mapper.name_to_model_part_index(name)
            x[idx].append((name, rank))
        return x

    def prepare_local_shard_infos(
        self,
        hf_key_n_rank: List[Tuple[str, int]],
        global_rank: int,
    ) -> Dict[str, Any]:
        """
        Prepare local shard information for the given parameters based on the parallelism configuration.
        :param hf_key_n_rank: A list of tuples containing the parameter names and their shape ranks.
        :param global_rank: The global rank to prepare local shard information for.
        :return: A dictionary containing the local shard information for each parameter of that rank.
        """
        x = self._cluster_params_by_model_part(hf_key_n_rank)
        insts = {}
        for model_index, p in enumerate(x):
            insts.update(
                self.mapper_group[model_index].local_shard_info_for_params(
                    p, global_rank
                )
            )
        return insts


class ParallelizedShardMapper:
    """
    A class to manage the parallelized shard mapping for policy and rollout models.
    This class is responsible for gathering shard information for policy and rollout, generating send and receive instructions,
    and managing the parallelism configuration for both policy and rollout models.
    It uses multiprocessing to handle the unpacking of shard information in parallel.
    """

    def __init__(self, config: CosmosConfig, max_processes: Optional[int] = None):
        """
        Initialize the ParallelizedShardMapper with empty shard info lists.
        ParallelizedShardMapper is used to gather the shard information for policy and rollout then generate the send and receive instructions for policy and rollout.
        :param config: The CosmosConfig instance containing the configuration for the parallelism.
        :param max_processes: The maximum number of processes to use for parallel processing.
        """
        self.policy_all_rank_shard_infos: Optional[List[Dict[str, Any]]] = None
        self.rollout_all_rank_shard_infos: Optional[List[Dict[str, Any]]] = None

        self.send_insts_for_policy: Optional[
            List[List[WeightSyncInstructionsGroup]]
        ] = None  # List of instructions of different ranks and for each rank it contains a list of instructions for each parameter group.
        self.recv_insts_for_rollout: Optional[
            List[List[WeightSyncInstructionsGroup]]
        ] = None  # List of instructions of different ranks and for each rank it contains a list of instructions for each parameter group.
        self.config = config
        self.param_to_param_groups = {}
        self.param_groups = set()
        self.policy_raw_info_bytes = None
        self.rollout_raw_info_bytes = None
        self.max_processes = max_processes
        self.scheme_generation_done = asyncio.Event()
        self.scheme_generation_done.clear()

    @classmethod
    def get_instance(
        cls, config: CosmosConfig = None, max_processes: Optional[int] = None
    ) -> "ParallelizedShardMapper":
        """
        Get the singleton instance of ParallelizedShardMapper.
        :param config: The CosmosConfig instance containing the configuration for the parallelism.
        :param max_processes: The maximum number of processes to use for parallel processing.
        :return: The singleton instance of ParallelizedShardMapper.
        """
        if not hasattr(cls, "_instance"):
            cls._instance = cls(config, max_processes)
        return cls._instance

    @staticmethod
    def unpack_and_set_shard_infos(
        is_policy: bool = True,
    ):
        """
        Unpack and set the shard information for policy and rollout base on raw bytes received.
        :param is_policy: A boolean indicating if the unpacking is for policy or rollout.
        This method is called to extrac the shard information for policy and rollout after they are unpacked from the bytes.
        """
        instance = ParallelizedShardMapper.get_instance()
        if is_policy:
            unpacked_data = msgpack.unpackb(
                instance.policy_raw_info_bytes, strict_map_key=False
            )
        else:
            unpacked_data = msgpack.unpackb(
                instance.rollout_raw_info_bytes, strict_map_key=False
            )
        return unpacked_data

    async def post_set_shard_infos(self):
        """
        Post-process the shard infos after they are set.
        This method generates the send and receive instructions for policy and rollout based on the shard information provided.
        It initializes the send and receive instruction lists for policy and rollout, respectively.
        """
        if (
            self.policy_raw_info_bytes is not None
            and self.rollout_raw_info_bytes is not None
            and self.send_insts_for_policy is None
            and self.recv_insts_for_rollout is None
        ):
            logger.debug(
                "[ParallelizedShardMapper] Start unpacking shard infos for both policy and rollout."
            )
            self.send_insts_for_policy = []
            self.recv_insts_for_rollout = []
            loop = asyncio.get_event_loop()
            ctx = multiprocessing.get_context("fork")
            with ProcessPoolExecutor(mp_context=ctx, max_workers=2) as pool:
                policy_unpack = loop.run_in_executor(
                    pool,
                    ParallelizedShardMapper.unpack_and_set_shard_infos,
                    True,
                )
                rollout_unpack = loop.run_in_executor(
                    pool,
                    ParallelizedShardMapper.unpack_and_set_shard_infos,
                    False,
                )
                policy_info = await policy_unpack
                rollout_info = await rollout_unpack
            self.policy_raw_info_bytes = None
            self.rollout_raw_info_bytes = None

            self.policy_all_rank_shard_infos = policy_info["shard_infos"]
            for group in policy_info["param_groups"]:
                for param_name in group:
                    self.param_to_param_groups[param_name] = group
                # Each group has already been sorted in replica.
                self.param_groups.add(tuple(group))
            self.sorted_params_all_rank_policy = policy_info["sorted_params"]
            self.rollout_all_rank_shard_infos = rollout_info["shard_infos"]
            for group in rollout_info["param_groups"]:
                for param_name in group:
                    self.param_to_param_groups[param_name] = group
                # Each group has already been sorted in replica.
                self.param_groups.add(tuple(group))
            self.sorted_params_all_rank_rollout = rollout_info["sorted_params"]
            logger.debug(
                "[ParallelizedShardMapper] Finished unpacking shard infos for both policy and rollout."
            )
            self.param_groups = sorted(self.param_groups, key=lambda x: x[0])
            policy_parallelism = ParallelDims.from_config_for_analysis(
                self.config.policy.parallelism,
                self.policy_world_size,
            )
            rollout_parallelism = ParallelDims.from_config_for_analysis(
                self.config.rollout.parallelism,
                self.rollout_world_size,
            )
            self.pp_rank_map_policy = [
                policy_parallelism.get_rank_in_dim("pp", p_rank)
                for p_rank in range(self.policy_world_size)
            ]
            self.pp_rank_map_rollout = [
                rollout_parallelism.get_rank_in_dim("pp", r_rank)
                for r_rank in range(self.rollout_world_size)
            ]
            self.send_insts_for_policy = [None for _ in range(self.policy_world_size)]
            self.recv_insts_for_rollout = [None for _ in range(self.rollout_world_size)]
            ctx = multiprocessing.get_context("fork")
            self.multiprocessing_pool = ProcessPoolExecutor(
                mp_context=ctx, max_workers=self.max_processes
            )
            self.policy_results = [
                loop.run_in_executor(
                    self.multiprocessing_pool,
                    ParallelizedShardMapper.policy_insts_generation_per_rank,
                    rank,
                )
                for rank in range(self.policy_world_size)
            ]
            self.rollout_results = [
                loop.run_in_executor(
                    self.multiprocessing_pool,
                    ParallelizedShardMapper.rollout_insts_generation_per_rank,
                    rank,
                )
                for rank in range(self.rollout_world_size)
            ]
            self.scheme_generation_done.set()
            logger.debug(
                "[ParallelizedShardMapper] Started generating send and receive instructions for policy and rollout asynchronously."
            )

    async def set_shard_infos_of_policy(
        self,
        policy_raw_info_bytes: bytes,
        n_gpus_per_policy: int,
    ):
        """
        Set the shard information for the policy.
        :param policy_raw_info_bytes: The raw bytes containing the shard information for the policy. include:
            shard_infos: A list of dictionaries containing shard info for each rank.
            sorted_params: A list of sorted parameter names for each rank.
            param_groups: A list of parameter groups for the policy. Each group contains multiple parameters with some connections such as from the same original param.
        :param n_gpus_per_policy: The number of GPUs per policy.
        """
        if self.policy_raw_info_bytes is None:
            self.policy_world_size = n_gpus_per_policy
            self.policy_raw_info_bytes = policy_raw_info_bytes
            util.create_async_task(self.post_set_shard_infos())

    async def set_shard_infos_of_rollout(
        self,
        rollout_raw_info_bytes: bytes,
        n_gpus_per_rollout: int,
    ):
        """
        Set the shard information for the rollout.
        :param rollout_raw_info_bytes: The raw bytes containing the shard information for the rollout. include:
            shard_infos: A list of dictionaries containing shard info for each rank.
            sorted_params: A list of sorted parameter names for each rank.
            param_groups: A list of parameter groups for the rollout. Each group contains multiple parameters with some connections such as from the same original param.
        :param n_gpus_per_rollout: The number of GPUs per rollout.
        """
        if self.rollout_raw_info_bytes is None:
            self.rollout_world_size = n_gpus_per_rollout
            self.rollout_raw_info_bytes = rollout_raw_info_bytes
            util.create_async_task(self.post_set_shard_infos())

    def generate_parallelized_shard_send_insts_for_policy(
        self, p_rank: int
    ) -> List[WeightSyncInstructionsGroup]:
        """
        Generate the send instructions for policy based on the shard information.
        :param p_rank: The rank of the policy to generate send instructions for.
        :return: A list of send instructions for policy.
        In the returned list, each element corresponds to a parameter group.
        Each parameter group contains a list of instructions for each parameter in a group represented by `WeightSyncInstructionsGroup`.
        One parameter group includes the parameters with some connections such as from the same original param.
        Correspondingly, each `WeightSyncInstructionsGroup` includes `WeightSyncInstructionsPerTensor` for each parameter in the group.
        Each instruction is a `WeightSyncInstruction` containing the parameter name and the corresponding sharded tensor split strategies.
        """
        sorted_params = self.sorted_params_all_rank_policy[
            self.pp_rank_map_policy[p_rank]
        ]
        policy_shard_dicts = self.policy_all_rank_shard_infos[p_rank]
        policy_to_rollout_insts = []
        name_in_group = set()
        for dest_name in sorted_params:
            insts_for_group = []
            if dest_name not in policy_shard_dicts:
                continue
            if dest_name in self.param_to_param_groups:
                name_in_group.add(dest_name)
                continue
            p_info = policy_shard_dicts[dest_name]
            insts_for_param_name = []
            shard_info = {k: DimSliceInfo.from_dict(v) for k, v in p_info.items()}
            p_dup_ranks = self.get_dup_ranks_for_policy(p_rank, dest_name)
            for r_rank, r_infos in enumerate(self.rollout_all_rank_shard_infos):
                if dest_name not in r_infos:
                    continue
                r_info = r_infos[dest_name]
                r_shard_info = {k: DimSliceInfo.from_dict(v) for k, v in r_info.items()}
                r_dup_ranks = self.get_dup_ranks_for_rollout(r_rank, dest_name)

                all_dims = shard_info.keys() | r_shard_info.keys()

                p_tensor_split_strategys = {}
                r_tensor_split_strategys = {}
                for d in all_dims:
                    p_tensor_split_strategy, r_tensor_split_strategy = (
                        ParallelTopoMapper.tensor_overlap_info_at_dim(
                            shard_info, r_shard_info, d
                        )
                    )
                    if p_tensor_split_strategy is None:
                        assert r_tensor_split_strategy is None
                        p_tensor_split_strategys = None
                        break
                    p_tensor_split_strategys[d] = p_tensor_split_strategy.__dict__
                    r_tensor_split_strategys[d] = r_tensor_split_strategy.__dict__
                if p_tensor_split_strategys is None:
                    continue
                else:
                    p_assignments, r_assignments = (
                        ParallelTopoMapper.policy_to_rollout_assign(
                            p_dup_ranks, r_dup_ranks, p_rank, r_rank
                        )
                    )
                    for r in p_assignments:
                        insts_for_param_name.append(
                            WeightSyncInstruction(
                                p_rank, r, p_tensor_split_strategys
                            ).__dict__
                        )
            insts_for_group.append(
                WeightSyncInstructionsPerParam(dest_name, insts_for_param_name).__dict__
            )
            policy_to_rollout_insts.append(
                WeightSyncInstructionsGroup(insts_for_group).__dict__
            )
            if not insts_for_param_name:
                logger.warning(
                    f"No send instructions generated for parameter {dest_name} in policy rank {p_rank}."
                )
        for group in self.param_groups:
            insts_for_group = []
            for dest_name in group:
                if dest_name not in policy_shard_dicts:
                    continue
                if dest_name in name_in_group:
                    name_in_group.remove(dest_name)
                p_info = policy_shard_dicts[dest_name]
                insts_for_param_name = []
                shard_info = {k: DimSliceInfo.from_dict(v) for k, v in p_info.items()}
                p_dup_ranks = self.get_dup_ranks_for_policy(p_rank, dest_name)
                for r_rank, r_infos in enumerate(self.rollout_all_rank_shard_infos):
                    if dest_name not in r_infos:
                        continue
                    r_info = r_infos[dest_name]
                    r_shard_info = {
                        k: DimSliceInfo.from_dict(v) for k, v in r_info.items()
                    }
                    r_dup_ranks = self.get_dup_ranks_for_rollout(r_rank, dest_name)

                    all_dims = shard_info.keys() | r_shard_info.keys()

                    p_tensor_split_strategys = {}
                    for d in all_dims:
                        p_tensor_split_strategy, r_tensor_split_strategy = (
                            ParallelTopoMapper.tensor_overlap_info_at_dim(
                                shard_info, r_shard_info, d
                            )
                        )
                        if p_tensor_split_strategy is None:
                            assert r_tensor_split_strategy is None
                            p_tensor_split_strategys = None
                            break
                        p_tensor_split_strategys[d] = p_tensor_split_strategy.__dict__
                    if p_tensor_split_strategys is None:
                        continue
                    else:
                        p_assignments, _ = ParallelTopoMapper.policy_to_rollout_assign(
                            p_dup_ranks, r_dup_ranks, p_rank, r_rank
                        )
                        for r in p_assignments:
                            insts_for_param_name.append(
                                WeightSyncInstruction(
                                    p_rank, r, p_tensor_split_strategys
                                ).__dict__
                            )
                insts_for_group.append(
                    WeightSyncInstructionsPerParam(
                        dest_name, insts_for_param_name
                    ).__dict__
                )
            policy_to_rollout_insts.append(
                WeightSyncInstructionsGroup(insts_for_group).__dict__
            )
        if len(name_in_group) > 0:
            logger.warning(
                f"No send instructions generated for parameters {name_in_group} in policy rank {p_rank}."
            )
        # Pack the instructions into msgpack format for efficient serialization.
        return msgpack.packb(policy_to_rollout_insts)

    def get_dup_ranks_for_policy(
        self,
        p_rank: int,
        name: str,
    ) -> List[List[Dict[str, Any]]]:
        """
        Get the duplicate ranks for a param in policy of the given rank.
        :param p_rank: The rank of the policy to get duplicate ranks for.
        :param name: The name of the parameter to get duplicate ranks for.
        :return: A list of duplicate ranks for policy.
        """
        if name in self.policy_all_rank_shard_infos[p_rank]:
            p_info = self.policy_all_rank_shard_infos[p_rank][name]
        else:
            return []

        ranks = []
        for p_rank, p_infos in enumerate(self.policy_all_rank_shard_infos):
            if name in p_infos:
                if p_infos[name] == p_info:
                    ranks.append(p_rank)
        return ranks

    def get_dup_ranks_for_rollout(
        self,
        p_rank: int,
        name: str,
    ) -> List[List[Dict[str, Any]]]:
        """
        Get the duplicate ranks for a param in rollout of the given rank.
        :param p_rank: The rank of the rollout to get duplicate ranks for.
        :param name: The name of the parameter to get duplicate ranks for.
        :return: A list of duplicate ranks for rollout.
        """
        if name in self.rollout_all_rank_shard_infos[p_rank]:
            p_info = self.rollout_all_rank_shard_infos[p_rank][name]
        else:
            return []

        ranks = []
        for p_rank, p_infos in enumerate(self.rollout_all_rank_shard_infos):
            if name in p_infos:
                if p_infos[name] == p_info:
                    ranks.append(p_rank)
        return ranks

    def generate_parallelized_shard_recv_insts_for_rollout(
        self, r_rank: int
    ) -> List[WeightSyncInstructionsGroup]:
        """
        Generate the receive instructions for rollout based on the shard information.
        :param r_rank: The rank of the rollout to generate receive instructions for.
        :return: A list of receive instructions for rollout.
        In the returned list, each element corresponds to a parameter group.
        Each parameter group contains a list of instructions for each parameter in a group represented by `WeightSyncInstructionsGroup`.
        One parameter group includes the parameters with some connections such as from the same original param.
        Correspondingly, each `WeightSyncInstructionsGroup` includes `WeightSyncInstructionsPerTensor` for each parameter in the group.
        Each instruction is a `WeightSyncInstruction` containing the parameter name and the corresponding sharded tensor split strategies.
        """
        sorted_params = self.sorted_params_all_rank_rollout[
            self.pp_rank_map_rollout[r_rank]
        ]
        rollout_shard_dicts = self.rollout_all_rank_shard_infos[r_rank]
        rollout_from_policy_insts = []
        name_in_group = set()
        for dest_name in sorted_params:
            insts_for_group = []
            if dest_name not in rollout_shard_dicts:
                continue
            if dest_name in self.param_to_param_groups:
                name_in_group.add(dest_name)
                continue
            r_info = rollout_shard_dicts[dest_name]
            insts_for_param_name = []
            shard_info = {k: DimSliceInfo.from_dict(v) for k, v in r_info.items()}
            r_dup_ranks = self.get_dup_ranks_for_rollout(r_rank, dest_name)
            for p_rank, p_infos in enumerate(self.policy_all_rank_shard_infos):
                if dest_name not in p_infos:
                    continue
                p_info = p_infos[dest_name]
                p_shard_info = {k: DimSliceInfo.from_dict(v) for k, v in p_info.items()}
                p_dup_ranks = self.get_dup_ranks_for_policy(p_rank, dest_name)

                all_dims = p_shard_info.keys() | shard_info.keys()
                r_tensor_split_strategys = {}
                for d in all_dims:
                    p_tensor_split_strategy, r_tensor_split_strategy = (
                        ParallelTopoMapper.tensor_overlap_info_at_dim(
                            p_shard_info, shard_info, d
                        )
                    )
                    if r_tensor_split_strategy is None:
                        assert p_tensor_split_strategy is None
                        r_tensor_split_strategys = None
                        break
                    r_tensor_split_strategys[d] = r_tensor_split_strategy.__dict__
                if r_tensor_split_strategys is None:
                    continue
                else:
                    _, r_assignments = ParallelTopoMapper.policy_to_rollout_assign(
                        p_dup_ranks, r_dup_ranks, p_rank, r_rank
                    )
                    for p in r_assignments:
                        insts_for_param_name.append(
                            WeightSyncInstruction(
                                p, r_rank, r_tensor_split_strategys
                            ).__dict__
                        )
            if insts_for_param_name:
                insts_for_group.append(
                    WeightSyncInstructionsPerParam(
                        dest_name, insts_for_param_name
                    ).__dict__
                )
            if insts_for_group:
                rollout_from_policy_insts.append(
                    WeightSyncInstructionsGroup(insts_for_group).__dict__
                )
            else:
                raise ValueError(
                    f"No recv instructions generated for parameter {dest_name} in rollout rank {r_rank}."
                )
        for group in self.param_groups:
            insts_for_group = []
            for dest_name in group:
                if dest_name not in rollout_shard_dicts:
                    continue
                if dest_name in name_in_group:
                    name_in_group.remove(dest_name)
                r_info = rollout_shard_dicts[dest_name]
                insts_for_param_name = []
                shard_info = {k: DimSliceInfo.from_dict(v) for k, v in r_info.items()}
                r_dup_ranks = self.get_dup_ranks_for_rollout(r_rank, dest_name)
                for p_rank, p_infos in enumerate(self.policy_all_rank_shard_infos):
                    if dest_name not in p_infos:
                        continue
                    p_info = p_infos[dest_name]
                    p_shard_info = {
                        k: DimSliceInfo.from_dict(v) for k, v in p_info.items()
                    }
                    p_dup_ranks = self.get_dup_ranks_for_policy(p_rank, dest_name)

                    all_dims = p_shard_info.keys() | shard_info.keys()
                    r_tensor_split_strategys = {}
                    for d in all_dims:
                        p_tensor_split_strategy, r_tensor_split_strategy = (
                            ParallelTopoMapper.tensor_overlap_info_at_dim(
                                p_shard_info, shard_info, d
                            )
                        )
                        if r_tensor_split_strategy is None:
                            assert p_tensor_split_strategy is None
                            r_tensor_split_strategys = None
                            break
                        r_tensor_split_strategys[d] = r_tensor_split_strategy.__dict__
                    if r_tensor_split_strategys is None:
                        continue
                    else:
                        _, r_assignments = ParallelTopoMapper.policy_to_rollout_assign(
                            p_dup_ranks, r_dup_ranks, p_rank, r_rank
                        )
                        for p in r_assignments:
                            insts_for_param_name.append(
                                WeightSyncInstruction(
                                    p, r_rank, r_tensor_split_strategys
                                ).__dict__
                            )
                if insts_for_param_name:
                    insts_for_group.append(
                        WeightSyncInstructionsPerParam(
                            dest_name, insts_for_param_name
                        ).__dict__
                    )
            if insts_for_group:
                rollout_from_policy_insts.append(
                    WeightSyncInstructionsGroup(insts_for_group).__dict__
                )
        assert (
            len(name_in_group) == 0
        ), f"No recv instructions generated for parameters {name_in_group} in rollout rank {r_rank}."
        # Pack the instructions into msgpack format for efficient serialization.
        return msgpack.packb(rollout_from_policy_insts)

    @staticmethod
    def policy_insts_generation_per_rank(
        rank: int,
    ) -> Tuple[List[WeightSyncInstructionsGroup], List[Dict[str, Any]]]:
        """
        Generate the send instructions for policy in parallel for a given rank.
        This method is called in parallel for each rank to generate the send instructions.
        :param rank: The rank of the current process.
        :return: The generated send instructions for policy of that rank and metadata used for the rollout from policy instructions.
        """
        logger.debug(
            f"[ParallelizedShardMapper] Start generating send instructions for rank {rank}"
        )
        mapper = ParallelizedShardMapper.get_instance()
        if rank < mapper.policy_world_size:
            send_insts = mapper.generate_parallelized_shard_send_insts_for_policy(rank)
        else:
            send_insts = None
        logger.debug(
            f"[ParallelizedShardMapper] Finished generating send instructions for rank {rank}"
        )
        return send_insts

    @staticmethod
    def rollout_insts_generation_per_rank(
        rank: int,
    ) -> List[WeightSyncInstructionsGroup]:
        """
        Generate the receive instructions for rollout in parallel for a given rank.
        This method is called in parallel for each rank to generate the receive instructions.
        :param rank: The rank of the current process.
        :return: The generated receive instructions for rollout of that rank.
        """
        logger.debug(
            f"[ParallelizedShardMapper] Start generating receive instructions for rank {rank}"
        )
        mapper = ParallelizedShardMapper.get_instance()
        if rank < mapper.rollout_world_size:
            recv_insts = mapper.generate_parallelized_shard_recv_insts_for_rollout(rank)
        else:
            recv_insts = None
        logger.debug(
            f"[ParallelizedShardMapper] Finished generating receive instructions for rank {rank}"
        )
        return recv_insts

    async def get_send_insts_for_policy(
        self, rank: int
    ) -> List[WeightSyncInstructionsGroup]:
        """
        Get the send instructions for policy of the given rank.
        :param rank: The rank of the policy to get send instructions for.
        :return: A list of send instructions for policy.
        """
        if self.send_insts_for_policy[rank] is None:
            self.send_insts_for_policy[rank] = await self.policy_results[rank]
        return self.send_insts_for_policy[rank]

    async def get_recv_insts_for_rollout(
        self, rank: int
    ) -> List[WeightSyncInstructionsGroup]:
        """
        Get the receive instructions for rollout of the given rank.
        :param rank: The rank of the rollout to get receive instructions for.
        :return: A list of receive instructions for rollout.
        """
        if self.recv_insts_for_rollout[rank] is None:
            self.recv_insts_for_rollout[rank] = await self.rollout_results[rank]
        return self.recv_insts_for_rollout[rank]



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import torch
import time
import os

os.environ["COSMOS_HEARTBEAT_TIMEOUT"] = "100000"
import sys
import subprocess
import torch.distributed as dist
import threading
from functools import partial
import requests
import atexit
from queue import Queue, Empty

from cosmos_rl.utils.logging import logger
from cosmos_rl.utils import constant
from cosmos_rl.utils.network_util import make_request_with_retry, get_local_ip
from cosmos_rl.utils.distributed import HighAvailabilitylNccl
from cosmos_rl.dispatcher.command import Command, BuildMeshCommand
from cosmos_rl.policy.config import Config
from cosmos_rl.utils.distributed import get_controller_metadata
from cosmos_rl.dispatcher.protocol import MESH_NAMES
from cosmos_rl.comm.base import CommMixin
from cosmos_rl.utils.parallelism import ParallelDims
from cosmos_rl.utils.api_suffix import (
    COSMOS_API_REGISTER_SUFFIX,
    COSMOS_API_UNREGISTER_SUFFIX,
)


WORK_DIR = f"/tmp/{os.path.basename(__file__)}"
CTRL_PORT = 8010


os.environ["COSMOS_CONTROLLER_HOST"] = f"localhost:{CTRL_PORT}"
os.environ["COSMOS_NCCL_TIMEOUT_MS"] = "30000"
os.environ["COSMOS_LOG_LEVEL"] = "DEBUG"

# os.environ["NCCL_DEBUG"] = "INFO"


def write_train_config():
    os.makedirs(WORK_DIR, exist_ok=True)

    config = f"""
redis = "12808"

[train]
resume = "False"
epoch = 1
output_dir = "{WORK_DIR}"

[logging]
logger = ['console', 'wandb']
project_name = "cosmos_rl"
experiment_name = "None"

[train.train_policy]
type = "grpo"
dataset.name = "JiaxinTsao/math_examples"
prompt_column_name = "prompt"
response_column_name = "result"
dataset.split = "train"
reward_function = "boxed_math"
temperature = 0.9
epsilon_low = 0.2
epsilon_high = 0.2
kl_beta = 0.0
mu_iterations = 1
min_filter_prefix_tokens = 1

[policy.parallelism]
n_init_replicas = 1
tp_size = 1
cp_size = 1
dp_shard_size = 1
pp_size = 1
dp_replicate_size = 1
"""
    cfg_file_path = os.path.join(WORK_DIR, "train_config.toml")
    with open(cfg_file_path, "w") as f:
        f.write(config)
    return cfg_file_path


def launch_controller(config: str):
    """
    Launch the controller process.
    """
    logger.info("launch controller")
    env = os.environ.copy()
    env["COSMOS_ROLE"] = "Controller"
    p = subprocess.Popen(
        "python -m cosmos_rl.dispatcher.run_web_panel "
        f"--port {CTRL_PORT} --config {config}",
        shell=True,
        stdout=sys.stdout,
        stderr=sys.stderr,
        text=True,
        env=env,
    )

    return [p]


class TestHANccl(CommMixin):
    def __init__(self, ctrl_ip: str, ctrl_port: int, config: Config):
        self.config = config
        self.replica_name = f"HANCCL-{dist.get_rank()}"
        self.global_rank = (
            0  # here we assume every replica only has one gpu, so the global_rank is 0
        )
        self.replica_rank = dist.get_rank()
        self.remote_hosts = [f"http://{os.environ['COSMOS_CONTROLLER_HOST']}"]
        self.role = "POLICY"
        self._shutdown_event = threading.Event()
        self._is_registered = False

        # override those fields
        self.remote_ips = ctrl_ip
        self.remote_port = ctrl_port
        self.parallel_dims = ParallelDims.from_config_for_analysis(
            config.policy.parallelism, world_size=1
        )
        self.parallel_dims.build_mesh(device_type="cuda")

        # init redis
        self.init_redis()

        # start heartbeat thread
        # self.heartbeat_thread = self.start_heartbeat(
        #     self._shutdown_event,
        # )

        self.fetch_command_thread = threading.Thread(
            target=self.run_fetch_command,
            daemon=True,
        )
        self.fetch_command_thread.start()
        self.command_queue = Queue()

    def __get_alternative_urls(self, suffix: str):
        # Get the alternative URLs for the given suffix
        urls = []
        for remote_host in self.remote_hosts:
            urls.append(f"{remote_host}/{suffix}")
        return urls

    def register_to_controller(self):
        if self._is_registered:
            return

        host_info_tuple = get_local_ip()
        if host_info_tuple is None:
            raise Exception("Failed to get local IP address")
        host_ip, host_name = host_info_tuple
        make_request_with_retry(
            partial(
                requests.post,
                json={
                    "replica_name": self.replica_name,
                    "role": self.role,
                    "mesh_names": MESH_NAMES,
                    "ranks": [0 for _ in MESH_NAMES],
                    "group_size": [1 for _ in MESH_NAMES],
                    "global_rank": self.global_rank,
                    "host_ip": host_ip,
                    "host_name": host_name,
                },
            ),
            self.__get_alternative_urls(COSMOS_API_REGISTER_SUFFIX),
            max_retries=constant.COSMOS_HTTP_RETRY_CONFIG.max_retries,
        )
        self._is_registered = True
        logger.info(f"register to controller: {self.replica_name}")

    def unregister_from_controller(self):
        if not self._is_registered:
            return

        make_request_with_retry(
            partial(
                requests.post,
                json={"replica_name": self.replica_name},
            ),
            self.__get_alternative_urls(COSMOS_API_UNREGISTER_SUFFIX),
            max_retries=constant.COSMOS_HTTP_RETRY_CONFIG.max_retries,
        )
        self._is_registered = False
        logger.info(f"unregister from controller: {self.replica_name}")

    def run_fetch_command(self):
        while not self._shutdown_event.is_set():
            if not self._is_registered:
                time.sleep(0.1)
                continue
            raw_cmds = self.redis_controller.subscribe_command(self.replica_name)

            for cmd in raw_cmds:
                self.command_queue.put(Command.depack(cmd))
            time.sleep(0.1)
        logger.info(f"fetch command thread is stopped: {self.replica_name}")

    def fetch_command(self, block: bool = True) -> list[Command]:
        if not self._is_registered:
            return []

        try:
            cmds = [self.command_queue.get(block=block)]
            while not self.command_queue.empty():
                cmds.append(self.command_queue.get())
        except Empty:
            return []
        return cmds

    def test_comm_auto_rebuild_normal(self):
        logger.info(f"run test_comm_auto_rebuild {self.replica_name}")

        # 0. register to controller, let controller trigger the buildmesh command
        self.register_to_controller()

        # 1. create a comm
        comm = HighAvailabilitylNccl(
            replica_name=self.replica_name,
            global_rank=self.global_rank,
            controller_hosts=self.remote_hosts,
        )

        # wait all ranks fetch latest command from controller
        dist.barrier()
        time.sleep(2)

        query_cmd = self.fetch_command()
        for cmd in query_cmd:
            if isinstance(cmd, BuildMeshCommand):
                logger.info(
                    f"  ({self.replica_name}) push build mesh command: {cmd.replica_name_to_rank}"
                )
                comm.push_cmd(cmd)

        # 2. wait for the comm to be ready
        comm.wait_comm_ready()
        assert (
            comm.world_size() == dist.get_world_size()
        ), f"world size should be {dist.get_world_size()}"
        comm.destroy_nccl_comm()
        comm.shutdown()
        logger.info("  === normal case, passed")

    def test_comm_auto_rebuild_intiative_scale_down(self):
        logger.info(
            f"run test_comm_auto_rebuild_intiative_scale_down {self.replica_name}"
        )

        # 0. register to controller, let controller trigger the buildmesh command
        self.register_to_controller()

        # 1. create a comm
        comm = HighAvailabilitylNccl(
            replica_name=self.replica_name,
            global_rank=self.global_rank,
            controller_hosts=self.remote_hosts,
        )

        # wait all ranks fetch latest command from controller
        dist.barrier()
        time.sleep(2)

        # 2. test intiative scale down
        if self.replica_rank == 1:
            # use unregister to trigger the buildmesh command
            self.unregister_from_controller()
        else:
            # other rank do scale down buildmesh
            while True:
                # wait until controller trigger the buildmesh command
                cmds = self.fetch_command()
                cmds = [cmd for cmd in cmds if isinstance(cmd, BuildMeshCommand)]
                if len(cmds) == 0:
                    time.sleep(0.1)
                    continue
                for cmd in cmds:
                    comm.push_cmd(cmd)
                break
            comm.wait_comm_ready()
            assert (
                comm.world_size() == dist.get_world_size() - 1
            ), f"world size should be {dist.get_world_size() - 1}, actual {comm.world_size()}"

        comm.destroy_nccl_comm()
        comm.shutdown()
        logger.info("  === intiative scale down, passed")

    def test_comm_auto_rebuild_timeout_scale_down(self):
        logger.info(
            f"run test_comm_auto_rebuild_intiative_scale_down_and_scale_up {self.replica_name}"
        )

        # 0. register to controller, let controller trigger the buildmesh command
        self.register_to_controller()

        # 1. create a comm
        comm = HighAvailabilitylNccl(
            replica_name=self.replica_name,
            global_rank=self.global_rank,
            controller_hosts=self.remote_hosts,
        )

        dist.barrier()
        world_size = dist.get_world_size()
        # 2. trigger buildmesh command over all ranks
        if self.replica_rank == 1:
            self.unregister_from_controller()
            return
        else:
            time.sleep(5)

        cmds = self.fetch_command()
        cmds = [cmd for cmd in cmds if isinstance(cmd, BuildMeshCommand)]
        assert (
            len(cmds) > 0
        ), f"  replica_rank {self.replica_rank} should have at least one buildmesh command"

        # 3. test build mesh timeout
        # let rank 1 exit, so that other ranks will timeout, and rebuild mesh in scale down mode
        logger.info(
            f"  replica_rank {self.replica_rank} prepare buildmesh timeout environment"
        )
        if self.replica_rank == 1:
            logger.info(
                "  replica_rank 1 exit, wait for other ranks to timeout and execute the build mesh command"
            )
        else:
            # monitor that all other ranks execute the build mesh command
            cmd = cmds[-1]
            assert (
                len(cmd.replica_name_to_rank) == world_size - 1
            ), f"buildmesh command should be {world_size - 1}"
            logger.info(
                f"  replica_rank {self.replica_rank} push the buildmesh command: {cmd.replica_name_to_rank}"
            )
            comm.push_cmd(cmd)

            retry_count = 0
            while retry_count < 100:
                # here we wait rebuild comm until nranks equals to the world size - 1
                if comm.is_ready():
                    break

                # query timeout commands and push to comm,
                cmds = self.fetch_command(block=False)
                cmds = [cmd for cmd in cmds if isinstance(cmd, BuildMeshCommand)]
                for cmd in cmds:
                    logger.info(
                        f"  replica_rank {self.replica_rank} retry, push cmd: {cmd.replica_name_to_rank}"
                    )
                    comm.push_cmd(cmd)

                time.sleep(5)
                retry_count += 1

            assert comm.is_ready(), "comm is not ready"
            assert (
                comm.world_size() == world_size - 1
            ), f"world size should be {world_size - 1}, actual {comm.world_size()}"

        # finally, shutdown the comm
        comm.destroy_nccl_comm()
        comm.shutdown()
        logger.info(" === test_comm_auto_rebuild passed")

    def test_allreduce_timeout_retry(self):
        """
        Run the NCCL test.
        """
        logger.info("run nccl test")
        pass

    def test_send_recv_timeout_retry():
        """
        Run the NCCL test.
        """
        logger.info("run nccl test")
        pass


def cleanup():
    """
    Cleanup function to kill all processes.
    """
    # try kill created processes
    login_user = "root"
    try:
        login_user = os.getlogin()
    except Exception:
        pass
    subprocess.run(f"pkill -u {login_user} -f 'redis'", shell=True)
    subprocess.run(f"pkill -u {login_user} -f 'dispatcher.run_web_panel'", shell=True)
    # subprocess.run(f"pkill -u {login_user} -f 'torchrun'", shell=True)


atexit.register(cleanup)


def main():
    # 1. init process group in this node
    local_rank = int(os.environ["LOCAL_RANK"])
    torch.cuda.set_device(local_rank)
    dist.init_process_group(backend="cuda:nccl,cpu:gloo")
    rank = dist.get_rank()

    # 2. only rank 0 launch the controller
    ctrl_hdl = []
    if rank == 0:
        cfg_path = write_train_config()
        ctrl_hdl.extend(launch_controller(cfg_path))

    # wait for controller to be ready
    time.sleep(10)

    # 3. init the tester
    ctrl_ip, ctrl_port, metadata = get_controller_metadata()
    cosmos_config = Config.from_dict(metadata["config"])
    tester = TestHANccl(ctrl_ip, ctrl_port, cosmos_config)
    dist.barrier()

    # 4. Run the test
    tester.test_comm_auto_rebuild_normal()
    dist.barrier()

    tester.test_comm_auto_rebuild_intiative_scale_down()
    dist.barrier()

    tester.test_comm_auto_rebuild_timeout_scale_down()
    dist.barrier()

    # finally, wait for all ranks to finish the test
    for hdl in ctrl_hdl:
        hdl.kill()
    dist.destroy_process_group()


if __name__ == "__main__":
    # let run this test in distributed mode
    if os.environ.get("RECURSIVE_ENTRYPOINT") is None:
        # n_gpu = torch.cuda.device_count()
        n_gpu = 4
        command = [
            "torchrun",
            "--nnodes",
            "1",
            "--nproc_per_node",
            str(n_gpu),
            os.path.abspath(__file__),
        ]
        env = os.environ.copy()
        env["RECURSIVE_ENTRYPOINT"] = "1"
        subprocess.run(command, env=env)
    else:
        main()
